#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Demonstrate usage of LFPy.Network with network of ball-and-stick type
morphologies with active HH channels inserted in the somas and passive-leak
channels distributed throughout the apical dendrite. The corresponding
morphology and template specifications are in the files BallAndStick.hoc and
BallAndStickTemplate.hoc.

Execution (w. MPI):

    mpirun -np 2 python example_network.py

Copyright (C) 2017 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""
# import modules:
import os
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
import scipy.signal as ss
import scipy.stats as st
from mpi4py import MPI
import neuron
from LFPy import NetworkCell, Network, Synapse, RecExtElectrode

# set up MPI variables:
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

# avoid same sequence of random numbers from numpy and neuron on each RANK,
# e.g., in order to draw unique cell and synapse locations and random synapse
# activation times
GLOBALSEED = 1234
np.random.seed(GLOBALSEED + RANK)

################################################################################
# Function declarations
################################################################################

def decimate(x, q=10, n=4, k=0.8, filterfun=ss.cheby1):
    """
    scipy.signal.decimate like downsampling using filtfilt instead of lfilter,
    and filter coeffs from butterworth or chebyshev type 1.

    Parameters
    ----------
    x : ndarray
        Array to be downsampled along last axis.
    q : int
        Downsampling factor.
    n : int
        Filter order.
    k : float
        Aliasing filter critical frequency Wn will be set as Wn=k/q.
    filterfun : function
        `scipy.signal.filter_design.cheby1` or
        `scipy.signal.filter_design.butter` function

    Returns
    -------
    ndarray
        Downsampled signal.

    """
    if not isinstance(q, int):
        raise TypeError("q must be an integer")

    if n is None:
        n = 1

    if filterfun == ss.butter:
        b, a = filterfun(n, k / q)
    elif filterfun == ss.cheby1:
        b, a = filterfun(n, 0.05, k / q)
    else:
        raise Exception('only ss.butter or ss.cheby1 supported')

    try:
        y = ss.filtfilt(b, a, x)
    except: # Multidim array can only be processed at once for scipy >= 0.9.0
        y = []
        for data in x:
            y.append(ss.filtfilt(b, a, data))
        y = np.array(y)

    try:
        return y[:, ::q]
    except:
        return y[::q]

def remove_axis_junk(ax, lines=['right', 'top']):
    """remove chosen lines from plotting axis"""
    for loc, spine in ax.spines.items():
        if loc in lines:
            spine.set_color('none')
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

def draw_lineplot(
        ax, data, dt=0.1,
        T=(0, 200),
        scaling_factor=1.,
        vlimround=None,
        label='local',
        scalebar=True,
        unit='mV',
        ylabels=True,
        color='r',
        ztransform=True,
        filter_data=False,
        filterargs=dict(N=2, Wn=0.02, btype='lowpass')):
    """helper function to draw line plots"""
    tvec = np.arange(data.shape[1])*dt
    tinds = (tvec >= T[0]) & (tvec <= T[1])

    # apply temporal filter
    if filter_data:
        b, a = ss.butter(**filterargs)
        data = ss.filtfilt(b, a, data, axis=-1)

    #subtract mean in each channel
    if ztransform:
        dataT = data.T - data.mean(axis=1)
        data = dataT.T

    zvec = -np.arange(data.shape[0])
    vlim = abs(data[:, tinds]).max()
    if vlimround is None:
        vlimround = 2.**np.round(np.log2(vlim)) / scaling_factor
    else:
        pass

    yticklabels = []
    yticks = []

    for i, z in enumerate(zvec):
        if i == 0:
            ax.plot(tvec[tinds], data[i][tinds] / vlimround + z, lw=1,
                    rasterized=False, label=label, clip_on=False,
                    color=color)
        else:
            ax.plot(tvec[tinds], data[i][tinds] / vlimround + z, lw=1,
                    rasterized=False, clip_on=False,
                    color=color)
        yticklabels.append('ch. %i' % (i+1))
        yticks.append(z)

    if scalebar:
        ax.plot([tvec[-1], tvec[-1]],
                [-1, -2], lw=2, color='k', clip_on=False)
        ax.text(tvec[-1]+np.diff(T)*0.02, -1.5,
                '$2^{' + '{}'.format(np.log2(vlimround)
                                    ) + '}$ ' + '{0}'.format(unit),
                color='k', rotation='vertical',
                va='center')

    ax.axis(ax.axis('tight'))
    ax.yaxis.set_ticks(yticks)
    if ylabels:
        ax.yaxis.set_ticklabels(yticklabels)
        ax.set_ylabel('channel', labelpad=0.1)
    else:
        ax.yaxis.set_ticklabels([])
    remove_axis_junk(ax, lines=['right', 'top'])
    ax.set_xlabel(r't (ms)', labelpad=0.1)

    return vlimround


################################################################################
# Set up shared and population-specific parameters
################################################################################
# relative path for simulation output:
OUTPUTPATH = 'example_network_output'

# class NetworkCell parameters:
cellParameters = dict(
    morphology='BallAndStick.hoc',
    templatefile='BallAndStickTemplate.hoc',
    templatename='BallAndStickTemplate',
    templateargs=None,
    delete_sections=False,
)

# class NetworkPopulation parameters:
populationParameters = dict(
    Cell=NetworkCell,
    cell_args=cellParameters,
    pop_args=dict(
        radius=100.,
        loc=0.,
        scale=20.),
    rotation_args=dict(x=0., y=0.),
)

# class Network parameters:
networkParameters = dict(
    dt=2**-4,
    tstop=1200.,
    v_init=-65.,
    celsius=6.5,
    OUTPUTPATH=OUTPUTPATH
)

# class RecExtElectrode parameters:
electrodeParameters = dict(
    x=np.zeros(13),
    y=np.zeros(13),
    z=np.linspace(1000., -200., 13),
    N=np.array([[0., 1., 0.] for _ in range(13)]),
    r=5.,
    n=50,
    sigma=0.3,
    method="soma_as_point"
)

# method Network.simulate() parameters:
networkSimulationArguments = dict(
    rec_current_dipole_moment=True,
    rec_pop_contributions=True,
    to_memory=True,
    to_file=False
)

# population names, sizes and connection probability:
population_names = ['E', 'I']
population_sizes = [80, 20]
connectionProbability = [[0.1, 0.1], [0.1, 0.1]]

# synapse model. All corresponding parameters for weights,
# connection delays, multapses and layerwise positions are
# set up as shape (2, 2) nested lists for each possible
# connection on the form:
# [["E:E", "E:I"],
#  ["I:E", "I:I"]].
synapseModel = neuron.h.Exp2Syn
# synapse parameters
synapseParameters = [[dict(tau1=0.2, tau2=1.8, e=0.),
                      dict(tau1=0.2, tau2=1.8, e=0.)],
                     [dict(tau1=0.1, tau2=9.0, e=-80.),
                      dict(tau1=0.1, tau2=9.0, e=-80.)]]
# synapse max. conductance (function, mean, st.dev., min.):
weightFunction = np.random.normal
weightArguments = [[dict(loc=0.002, scale=0.0002),
                    dict(loc=0.002, scale=0.0002)],
                   [dict(loc=0.02, scale=0.002),
                    dict(loc=0.02, scale=0.002)]]
minweight = 0.
# conduction delay (function, mean, st.dev., min.):
delayFunction = np.random.normal
delayArguments = [[dict(loc=1.5, scale=0.3),
                   dict(loc=1.5, scale=0.3)],
                  [dict(loc=1.5, scale=0.3),
                   dict(loc=1.5, scale=0.3)]]
mindelay = 0.3
multapseFunction = np.random.normal
multapseArguments = [[dict(loc=2., scale=.5), dict(loc=2., scale=.5)],
                     [dict(loc=5., scale=1.), dict(loc=5., scale=1.)]]
# method NetworkCell.get_rand_idx_area_and_distribution_norm
# parameters for layerwise synapse positions:
synapsePositionArguments = [[dict(section=['soma', 'apic'],
                                  fun=[st.norm, st.norm],
                                  funargs=[dict(loc=500., scale=100.),
                                           dict(loc=500., scale=100.)],
                                  funweights=[0.5, 1.]
                                 ) for _ in range(2)],
                            [dict(section=['soma', 'apic'],
                                  fun=[st.norm, st.norm],
                                  funargs=[dict(loc=0., scale=100.),
                                           dict(loc=0., scale=100.)],
                                  funweights=[1., 0.5]
                                 ) for _ in range(2)]]

if __name__ == '__main__':
    ############################################################################
    # Main simulation
    ############################################################################
    # create directory for output:
    if not os.path.isdir(OUTPUTPATH):
        if RANK == 0:
            os.mkdir(OUTPUTPATH)
    COMM.Barrier()

    # instantiate Network:
    network = Network(**networkParameters)

    # create E and I populations:
    for name, size in zip(population_names, population_sizes):
        network.create_population(name=name, POP_SIZE=size,
                                  **populationParameters)


        # create excitatory background synaptic activity for each cell
        # with Poisson statistics
        for cell in network.populations[name].cells:
            idx = cell.get_rand_idx_area_norm(section='allsec', nidx=64)
            for i in idx:
                syn = Synapse(cell=cell, idx=i, syntype='Exp2Syn',
                              weight=0.002,
                              **dict(tau1=0.2, tau2=1.8, e=0.))
                syn.set_spike_times_w_netstim(interval=100.)


    # create connectivity matrices and connect populations:
    for i, pre in enumerate(population_names):
        for j, post in enumerate(population_names):
            # boolean connectivity matrix between pre- and post-synaptic neurons
            # in each population (postsynaptic on this RANK)
            connectivity = network.get_connectivity_rand(
                pre=pre, post=post,
                connprob=connectionProbability[i][j]
                )

            # connect network:
            (conncount, syncount) = network.connect(
                pre=pre, post=post,
                connectivity=connectivity,
                syntype=synapseModel,
                synparams=synapseParameters[i][j],
                weightfun=np.random.normal,
                weightargs=weightArguments[i][j],
                minweight=minweight,
                delayfun=delayFunction,
                delayargs=delayArguments[i][j],
                mindelay=mindelay,
                multapsefun=multapseFunction,
                multapseargs=multapseArguments[i][j],
                syn_pos_args=synapsePositionArguments[i][j],
                )

    # set up extracellular recording device:
    electrode = RecExtElectrode(**electrodeParameters)

    # run simulation:
    SPIKES, OUTPUT, DIPOLEMOMENT = network.simulate(
        electrode=electrode,
        **networkSimulationArguments
    )

    # collect somatic potentials across all RANKs to RANK 0:
    if RANK == 0:
        somavs = []
        for i, name in enumerate(population_names):
            somavs.append([])
            somavs[i] += [cell.somav
                          for cell in network.populations[name].cells]
            for j in range(1, SIZE):
                somavs[i] += COMM.recv(source=j, tag=15)
    else:
        somavs = None
        for name in population_names:
            COMM.send([cell.somav for cell in network.populations[name].cells],
                      dest=0, tag=15)

    ############################################################################
    # Plot some output on RANK 0
    ############################################################################
    if RANK == 0:
        # spike raster
        fig, ax = plt.subplots(1, 1)
        for name, spts, gids in zip(population_names, SPIKES['times'], SPIKES['gids']):
            t = []
            g = []
            for spt, gid in zip(spts, gids):
                t = np.r_[t, spt]
                g = np.r_[g, np.zeros(spt.size)+gid]
            ax.plot(t[t >= 200], g[t >= 200], '|', label=name)
        ax.legend(loc=1)
        remove_axis_junk(ax, lines=['right', 'top'])
        ax.set_xlabel('t (ms)')
        ax.set_ylabel('gid')
        ax.set_title('spike raster')
        fig.savefig(os.path.join(OUTPUTPATH, 'spike_raster.pdf'),
                    bbox_inches='tight')
        plt.close(fig)


        # somatic potentials
        fig = plt.figure()
        gs = GridSpec(5, 1)
        ax = fig.add_subplot(gs[:4])
        draw_lineplot(ax, decimate(np.array(somavs)[0], q=16), dt=network.dt*16,
                      T=(200, 1200),
                      scaling_factor=1.,
                      vlimround=16,
                      label='E',
                      scalebar=True,
                      unit='mV',
                      ylabels=False,
                      color='C0',
                      ztransform=True
                     )
        ax.set_yticks([])
        ax.set_xticklabels([])
        ax.set_ylabel('E')
        ax.set_title('somatic potentials')
        ax.set_xlabel('')

        ax = fig.add_subplot(gs[4])
        draw_lineplot(ax, decimate(np.array(somavs)[1], q=16), dt=network.dt*16,
                      T=(200, 1200),
                      scaling_factor=1.,
                      vlimround=16,
                      label='I',
                      scalebar=True,
                      unit='mV',
                      ylabels=False,
                      color='C1',
                      ztransform=True
                     )
        ax.set_yticks([])
        ax.set_ylabel('I')

        fig.savefig(os.path.join(OUTPUTPATH, 'soma_potentials.pdf'),
                    bbox_inches='tight')
        plt.close(fig)


        # extracellular potentials, E and I contributions, sum
        fig, axes = plt.subplots(1, 3, figsize=(6.4, 4.8))
        fig.suptitle('extracellular potentials')
        for i, (ax, name, label) in enumerate(zip(axes, ['E', 'I', 'imem'],
                                                  ['E', 'I', 'sum'])):
            draw_lineplot(ax, decimate(OUTPUT[0][name], q=16), dt=network.dt*16,
                          T=(200, 1200),
                          scaling_factor=1.,
                          vlimround=None,
                          label=label,
                          scalebar=True,
                          unit='mV',
                          ylabels=True if i == 0 else False,
                          color='C{}'.format(i),
                          ztransform=True
                         )
            ax.set_title(label)
        fig.savefig(os.path.join(OUTPUTPATH, 'extracellular_potential.pdf'),
                    bbox_inches='tight')
        plt.close(fig)


        # current-dipole moments, E and I contributions, sum
        fig, axes = plt.subplots(3, 3, figsize=(6.4, 4.8))
        fig.subplots_adjust(wspace=0.45)
        fig.suptitle('current-dipole moments')
        for i, u in enumerate(['x', 'y', 'z']):
            for j, label in enumerate(['E', 'I', 'sum']):
                t = np.arange(DIPOLEMOMENT.shape[0])*network.dt
                inds = (t >= 200) & (t <= 1200)
                if label != 'sum':
                    axes[i, j].plot(t[inds][::16],
                                    decimate(DIPOLEMOMENT[label][inds, i],
                                             q=16),
                                    'C{}'.format(j))
                else:
                    axes[i, j].plot(t[inds][::16],
                                    decimate(DIPOLEMOMENT['E'][inds, i] +
                                             DIPOLEMOMENT['I'][inds, i], q=16),
                                    'C{}'.format(j))

                if j == 0:
                    axes[i, j].set_ylabel(r'$\mathbf{p}\cdot\mathbf{e}_{' +
                                          '{}'.format(u) +'}$ (nA$\mu$m)')
                if i == 0:
                    axes[i, j].set_title(label)
                if i != 2:
                    axes[i, j].set_xticklabels([])
                else:
                    axes[i, j].set_xlabel('t (ms)')
        fig.savefig(os.path.join(OUTPUTPATH, 'current_dipole_moment.pdf'),
                    bbox_inches='tight')
        plt.close(fig)


    # population illustration (per RANK)
    fig = plt.figure(figsize=(6.4, 4.8*2))
    ax = fig.add_subplot(111, projection='3d')
    ax.view_init(elev=5)
    ax.plot(electrode.x, electrode.y, electrode.z, 'ko', zorder=0)
    for i, (name, pop) in enumerate(network.populations.items()):
        for cell in pop.cells:
            c = 'C0' if name == 'E' else 'C1'
            ax.plot([cell.xstart[0], cell.xend[0]],
                    [cell.ystart[0], cell.yend[0]],
                    [cell.zstart[0], cell.zend[0]], c,
                    lw=5, zorder=-cell.xmid[0]-cell.ymid[0])
            ax.plot([cell.xstart[1], cell.xend[-1]],
                    [cell.ystart[1], cell.yend[-1]],
                    [cell.zstart[1], cell.zend[-1]], c,
                    lw=0.5, zorder=-cell.xmid[0]-cell.ymid[0])
    ax.set_xlabel('$x$ ($\mu$m)')
    ax.set_ylabel('$y$ ($\mu$m)')
    ax.set_zlabel('$z$ ($\mu$m)')
    ax.set_title('network populations')
    fig.savefig(os.path.join(OUTPUTPATH, 'population_RANK_{}.pdf'.format(RANK)),
                bbox_inches='tight')
    plt.close(fig)


    ############################################################################
    # customary cleanup of object references - the psection() function may not
    # write correct information if NEURON still has object references in memory,
    # even if Python references has been deleted. It will also allow the script
    # to be run in successive fashion.
    ############################################################################
    network.pc.gid_clear() # allows assigning new gids to threads
    electrode = None
    syn = None
    synapseModel = None
    for population in network.populations.values():
        for cell in population.cells:
            cell = None
        population.cells = None
        population = None
    pop = None
    network = None
    neuron.h('forall delete_section()')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""


import unittest
import os
import numpy as np
import LFPy
import neuron


class testMEG(unittest.TestCase):
    """
    test class LFPy.MEG
    """

    def test_MEG_00(self):
        '''test LFPy.MEG.calculate_H()'''
        current_dipole_moment = np.zeros((11, 3))
        current_dipole_moment[:, 0] += 1.
        dipole_location = np.zeros(3)
        sensor_locations = np.r_[np.eye(3), -np.eye(3)]

        gt = np.zeros((sensor_locations.shape[0],
                       current_dipole_moment.shape[0], 3))
        gt[1, :, 2] = 1./4/np.pi
        gt[2, :, 1] = -1./4/np.pi
        gt[4, :, 2] = -1./4/np.pi
        gt[5, :, 1] = 1./4/np.pi

        meg = LFPy.MEG(sensor_locations)
        np.testing.assert_equal(gt, meg.calculate_H(current_dipole_moment,
                                                    dipole_location))

    def test_MEG_01(self):
        '''test LFPy.MEG.calculate_H()'''
        current_dipole_moment = np.zeros((11, 3))
        current_dipole_moment[:, 1] += 1.
        dipole_location = np.zeros(3)
        sensor_locations = np.r_[np.eye(3), -np.eye(3)]

        gt = np.zeros((sensor_locations.shape[0],
                       current_dipole_moment.shape[0], 3))
        gt[0, :, 2] = -1./4/np.pi
        gt[2, :, 0] = 1./4/np.pi
        gt[3, :, 2] = 1./4/np.pi
        gt[5, :, 0] = -1./4/np.pi

        meg = LFPy.MEG(sensor_locations)
        np.testing.assert_equal(gt, meg.calculate_H(current_dipole_moment,
                                                    dipole_location))

    def test_MEG_02(self):
        '''test LFPy.MEG.calculate_H()'''
        current_dipole_moment = np.zeros((11, 3))
        current_dipole_moment[:, 2] += 1.
        dipole_location = np.zeros(3)
        sensor_locations = np.r_[np.eye(3), -np.eye(3)]

        # ground truth
        gt = np.zeros((sensor_locations.shape[0],
                       current_dipole_moment.shape[0], 3))
        gt[0, :, 1] = 1./4/np.pi
        gt[1, :, 0] = -1./4/np.pi
        gt[3, :, 1] = -1./4/np.pi
        gt[4, :, 0] = 1./4/np.pi

        meg = LFPy.MEG(sensor_locations)
        np.testing.assert_equal(gt, meg.calculate_H(current_dipole_moment,
                                                    dipole_location))

    def test_MEG_03(self):
        '''test LFPy.MEG.calculate_H()'''
        current_dipole_moment = np.zeros((1, 3))
        current_dipole_moment[:, 0] += 1.
        dipole_location = np.zeros(3)
        sensor_locations = np.r_[np.eye(3), -np.eye(3)]

        gt = np.zeros((sensor_locations.shape[0],
                       current_dipole_moment.shape[0], 3))
        gt[1, :, 2] = 1./4/np.pi
        gt[2, :, 1] = -1./4/np.pi
        gt[4, :, 2] = -1./4/np.pi
        gt[5, :, 1] = 1./4/np.pi

        meg = LFPy.MEG(sensor_locations)
        np.testing.assert_equal(gt, meg.calculate_H(current_dipole_moment,
                                                    dipole_location))

    def test_MEG_04(self):
        '''test LFPy.MEG.calculate_H()'''
        current_dipole_moment = np.zeros((1, 3))
        current_dipole_moment[:, 1] += 1.
        dipole_location = np.zeros(3)
        sensor_locations = np.r_[np.eye(3), -np.eye(3)]

        gt = np.zeros((sensor_locations.shape[0],
                       current_dipole_moment.shape[0], 3))
        gt[0, :, 2] = -1./4/np.pi
        gt[2, :, 0] = 1./4/np.pi
        gt[3, :, 2] = 1./4/np.pi
        gt[5, :, 0] = -1./4/np.pi

        meg = LFPy.MEG(sensor_locations)
        np.testing.assert_equal(gt, meg.calculate_H(current_dipole_moment,
                                                    dipole_location))

    def test_MEG_05(self):
        '''test LFPy.MEG.calculate_H()'''
        current_dipole_moment = np.zeros((1, 3))
        current_dipole_moment[:, 2] += 1.
        dipole_location = np.zeros(3)
        sensor_locations = np.r_[np.eye(3), -np.eye(3)]

        gt = np.zeros((sensor_locations.shape[0],
                       current_dipole_moment.shape[0], 3))
        gt[0, :, 1] = 1./4/np.pi
        gt[1, :, 0] = -1./4/np.pi
        gt[3, :, 1] = -1./4/np.pi
        gt[4, :, 0] = 1./4/np.pi

        meg = LFPy.MEG(sensor_locations)
        np.testing.assert_equal(gt, meg.calculate_H(current_dipole_moment,
                                                    dipole_location))


class testFourSphereVolumeConductor(unittest.TestCase):
    """
    test class LFPy.FourSphereVolumeConductor
    """

    def test_rz_params(self):
        radii = [1., 2., 4., 8.]
        sigmas = [1., 2., 4., 8.]
        r_el = np.array([[1., 0., 7.]])
        fs = LFPy.FourSphereVolumeConductor(radii, sigmas, r_el)

        rz1 = np.array([0., 0., 0.])
        with np.testing.assert_raises(RuntimeError):
            fs._rz_params(rz1)
        rz2 = np.array([0., 0., 1.])
        with np.testing.assert_raises(RuntimeError):
            fs._rz_params(rz2)
        rz3 = np.array([0., 0., 1.2])
        with np.testing.assert_raises(RuntimeError):
            fs._rz_params(rz3)

    def test_check_params00(self):
        '''Test that invalid radius values raises RuntimeError'''
        radii1 = [-1., 2., 4., 8.]
        radii2 = [1., .5, 4., 8.]
        radii3 = [1., 2., 1.1, 8.]
        radii4 = [1., 2., 4., 1.]
        sigmas = [1., 2., 4., 8.]
        r_el = np.array([[0., 0., 1.5]])
        with np.testing.assert_raises(RuntimeError):
            LFPy.FourSphereVolumeConductor(radii1, sigmas, r_el)
        with np.testing.assert_raises(RuntimeError):
            LFPy.FourSphereVolumeConductor(radii2, sigmas, r_el)
        with np.testing.assert_raises(RuntimeError):
            LFPy.FourSphereVolumeConductor(radii3, sigmas, r_el)
        with np.testing.assert_raises(RuntimeError):
            LFPy.FourSphereVolumeConductor(radii4, sigmas, r_el)

    def test_check_params01(self):
        '''Test that Error is raised if invalid entries in sigmas'''
        radii = [1., 2., 4., 10.]
        sigmas1 = [1., 'str', 4., 8.]
        sigmas2 = [-1., 2., 4., 8.]
        sigmas3 = [1., 2., -4., 8.]
        sigmas4 = [1., 2., 4., -8.]
        r_el = np.array([[0., 0., 1.5]])
        with np.testing.assert_raises(ValueError):
            LFPy.FourSphereVolumeConductor(radii, sigmas1, r_el)
        with np.testing.assert_raises(RuntimeError):
            LFPy.FourSphereVolumeConductor(radii, sigmas2, r_el)
        with np.testing.assert_raises(RuntimeError):
            LFPy.FourSphereVolumeConductor(radii, sigmas3, r_el)
        with np.testing.assert_raises(RuntimeError):
            LFPy.FourSphereVolumeConductor(radii, sigmas4, r_el)

    def test_check_params02(self):
        '''Test that ValueError is raised if electrode outside head'''
        radii = [1., 2., 4., 10.]
        sigmas = [1., 2., 4., 8.]
        r_el1 = np.array([[0., 0., 15.]])
        r_el2 = np.array([[0., 0., 1.5], [12., 0., 0.]])
        with np.testing.assert_raises(ValueError):
            LFPy.FourSphereVolumeConductor(radii, sigmas, r_el1)
        with np.testing.assert_raises(ValueError):
            LFPy.FourSphereVolumeConductor(radii, sigmas, r_el2)

    def test_decompose_dipole01(self):
        '''Test radial and tangential parts of dipole sums to dipole'''
        P1 = np.array([[1., 1., 1.]])
        p_rad, p_tan = decompose_dipole(P1)
        np.testing.assert_equal(p_rad + p_tan, P1)

    def test_decompose_dipole02(self):
        '''Test radial and tangential parts of dipole sums to dipole'''
        radii = [88000, 90000, 95000, 100000]
        sigmas = [0.3, 1.5, 0.015, 0.3]
        ps = np.array([[ 1000., 0., 0.],
                       [-1000., 0., 0.],
                       [0.,  1000., 0.],
                       [0., -1000., 0.],
                       [0., 0.,  1000.],
                       [0., 0., -1000.],
                       [10., 20., 30.],
                       [-10., -20., -30.]])
        p_locs = np.array([[ 87000., 0., 0.],
                           [-87000., 0., 0.],
                           [0.,  87000., 0.],
                           [0., -87000., 0.],
                           [0., 0.,  87000.],
                           [0., 0., -87000.],
                           [80000., 2000., 3000.],
                           [-2000., -80000., -3000.]])
        el_locs = np.array([[90000., 5000., -5000.]])
        fs = LFPy.FourSphereVolumeConductor(radii, sigmas, el_locs)
        for p_loc in p_locs:
            fs._rz_params(p_loc)
            p_rads, p_tans = fs._decompose_dipole(ps)
            np.testing.assert_equal(p_rads + p_tans, ps)

    def test_rad_dipole(self):
        '''Test that radial part of decomposed dipole is correct'''
        P1 = np.array([[1., 1., 1.]])
        p_rad, p_tan = decompose_dipole(P1)
        np.testing.assert_equal(p_rad, np.array([[0., 0., 1.]]))

    def test_tan_dipole(self):
        '''Test that tangential part of decomposed dipole is correct'''
        P1 = np.array([[1., 1., 1.]])
        p_rad, p_tan = decompose_dipole(P1)
        np.testing.assert_equal(p_tan, np.array([[1., 1., 0.]]))

    def test_calc_theta(self):
        '''Test theta: angle between rz and r'''
        rz1 = np.array([0., 0., 10.])
        r_el = np.array([[0., 0., 90.], [0., 0., -90.],[0., 70., 0.], [0., -70., 0.], [0., 10., 10.]])
        fs = make_class_object(rz1, r_el)
        theta = fs.calc_theta()
        np.testing.assert_almost_equal(theta, np.array([0., np.pi, np.pi/2, np.pi/2, np.pi/4]))

    def test_calc_phi01(self):
        '''Test phi: azimuthal angle between rx and rxy'''
        rz1 = np.array([0., 0., 0.5])
        r_el = np.array([[0., 1., 0], [-1., -1., 1.],
                         [1., 1., 4.], [0., 0., 89.], [0., 0., -80.]])
        fs = make_class_object(rz1, r_el)
        P_tan = np.array([[0., 1., 0.], [1., 0., 0.], [0., 0., 0.]])
        phi = fs.calc_phi(P_tan)
        np.testing.assert_almost_equal(phi, np.array([[np.pi/2, np.pi, 0.],
                                                      [-3*np.pi/4, -np.pi/4, 0.],
                                                      [np.pi/4, 3*np.pi/4, 0.],
                                                      [0., 0., 0.],
                                                      [0., 0., 0.]]))

    def test_calc_phi02(self):
        '''Test phi: azimuthal angle between rx and rxy,
           check that theta is not NaN, due to round-off errors'''
        radii = [79000., 80000., 85000., 100000.]
        sigmas = [0.3, 0.015, 15, 0.3]
        rz = np.array([0., 0., 76500.])
        r_el = np.array([[1e-5, 0, 99999.],
                         [0, 0.000123, 99998.9],
                         [-5.59822325e3, -9.69640709e3, -9.93712111e4],
                         [99990., 0., 0.001]])

        fs = LFPy.FourSphereVolumeConductor(radii, sigmas, r_el)
        fs._rz_params(rz)

        P1 = np.array([[0., 0., 123456789.],
                       [0., 0., 0.05683939],
                       [89892340., 0., -123456789],
                       [0.00004, 0.002, .0987654321],
                       [0., 0., 0.05683939],
                       [0.0003, 0.001, 123456789.],
                       [1e-11, 1e-12, 1000.],
                       [1e-15, 0, 1000.]])
        p_rad, p_tan = fs._decompose_dipole(P1)
        phi = fs.calc_phi(p_tan)

        np.testing.assert_equal(np.isnan(phi).any(), False)

    def test_rad_sign(self):
        '''Test if radial dipole points inwards or outwards'''
        rz1 = np.array([0., 0., 70.])
        r_el = np.array([[0., 0., 90.]])
        fs = make_class_object(rz1, r_el)
        P1 = np.array([[0., 0., 1.], [0., 0., -2.]])
        s_vector = fs._sign_rad_dipole(P1)
        np.testing.assert_almost_equal(s_vector, np.array([1., -1.]))

    def test_calc_vn(self):
        '''test that calc_vn gives correct values'''
        n = 1
        fs = make_simple_class_object()
        v1 = fs._calc_vn(1)
        np.testing.assert_almost_equal(v1, -4.75)

    def test_calc_yn(self):
        '''test that calc_yn gives correct values'''
        n = 1
        fs = make_simple_class_object()
        y1 = fs._calc_yn(1)
        np.testing.assert_almost_equal(y1, -2.3875)

    def test_calc_zn(self):
        '''test that calc_zn gives correct values'''
        n = 1
        fs = make_simple_class_object()
        z1 = fs._calc_zn(1)
        np.testing.assert_almost_equal(z1, -2.16574585635359)

    def test_calc_potential(self):
        '''test comparison between four-sphere model and model for
        infinite homogeneous space
        when sigma is constant and r4 goes to infinity'''
        sigmas = [0.3, 0.3, 0.3+1e-16, 0.3]
        radii = [10., 20*1e6, 30.*1e6, 40.*1e6]
        rz = np.array([0., 0., 3.])
        p = np.array([[0., 0., 100.], [50., 50., 0.]])
        r_elec = np.array([[0., 0., 9.],
                           [0., 0., 15.],
                           [0., 0., 25.],
                           [0., 0., 40.],
                           [0., 9., 0.],
                           [0., 15., 0.],
                           [0., 25., 0.],
                           [0., 40., 0.]])
        four_s = LFPy.FourSphereVolumeConductor(radii, sigmas, r_elec)
        pots_4s = four_s.calc_potential(p, rz)
        inf_s = LFPy.InfiniteVolumeConductor(0.3)
        pots_inf = inf_s.get_dipole_potential(p, r_elec - rz)

        np.testing.assert_allclose(pots_4s, pots_inf, rtol=1e-6)

    def test_calc_potential01(self):
        '''test comparison between analytical 4S-model and FEM simulation'''
        # load data
        fem_sim = np.load(os.path.join(LFPy.__path__[0], 'test', 'fem_mix_dip.npz'))
        pot_fem = fem_sim['pot_fem'] # [µV]
        p = fem_sim['p'] # [nAµm]
        rz = fem_sim['rz'] # [µm]
        radii = fem_sim['radii'] # [µm]
        sigmas = fem_sim['sigmas'] # [S/cm]
        ele_coords = fem_sim['ele_coords'] # [µm]

        fs = LFPy.FourSphereVolumeConductor(radii, sigmas, ele_coords)
        k_mV_to_muV = 1e3
        pot_analytical = fs.calc_potential(p, rz).reshape((len(ele_coords),)).reshape(pot_fem.shape)*k_mV_to_muV
        global_error = np.abs(pot_analytical - pot_fem)/(np.max(np.abs(pot_fem)))
        np.testing.assert_array_less(global_error, 0.01)

    def test_calc_potential02(self):
        '''Test radial and tangential parts of dipole sums to dipole'''
        radii = [88000, 90000, 95000, 100000]
        sigmas = [0.3, 1.5, 0.015, 0.3]

        dips = np.array([[[ 1000., 0., 0.]],
                         [[-1000., 0., 0.]],
                         [[0.,  1000., 0.]],
                         [[0., -1000., 0.]],
                         [[0., 0.,  1000.]],
                         [[0., 0., -1000.]]])

        p_locs = np.array([[ 87000., 0., 0.],
                           [-87000., 0., 0.],
                           [0.,  87000., 0.],
                           [0., -87000., 0.],
                           [0., 0.,  87000],
                           [0., 0., -87000]])

        el_locs = np.array([[[ 99000., 0., 0.]],
                            [[-99000., 0., 0.]],
                            [[0.,  99000., 0.]],
                            [[0., -99000., 0.]],
                            [[0., 0.,  99000.]],
                            [[0., 0., -99000.]]])


        for i in range(len(p_locs)):
            fs = LFPy.FourSphereVolumeConductor(radii, sigmas, el_locs[i])
            phi = fs.calc_potential(dips[i], p_locs[i])
            if i == 0:
                phi0 = phi[0][0]
            else:
                np.testing.assert_equal(phi0, phi[0][0])


    def test_calc_potential_from_multi_dipoles00(self):
        """test comparison between multi-dipoles and single dipole approach"""
        neuron.h('forall delete_section()')
        soma = neuron.h.Section(name='soma')
        dend1 = neuron.h.Section(name='dend1')
        dend2 = neuron.h.Section(name='dend2')
        dend1.connect(soma(0.5), 0)
        dend2.connect(dend1(1.0), 0)
        morphology = neuron.h.SectionList()
        morphology.wholetree()
        radii = [300, 400, 500, 600]
        sigmas = [0.3, 1.5, 0.015, 0.3]
        electrode_locs = np.array([[0., 0., 290.],
                                   [10., 90., 300.],
                                   [-90, 50., 400.],
                                   [110.3, -100., 500.]])
        cell = cell_w_synapse_from_sections(morphology)
        cell.set_pos(x=0, y=0, z=100)
        t_point = [1,100,-1]

        MD_4s = LFPy.FourSphereVolumeConductor(radii, sigmas, electrode_locs)
        p, dipole_locs = cell.get_multi_current_dipole_moments(t_point)
        Np, Nt, Nd = p.shape
        Ne = electrode_locs.shape[0]
        pot_MD = MD_4s.calc_potential_from_multi_dipoles(cell, t_point)

        pot_sum = np.zeros((Ne, Nt))
        for i in range(Np):
            dip = p[i]
            dip_loc = dipole_locs[i]
            fs = LFPy.FourSphereVolumeConductor(radii, sigmas, electrode_locs)
            pot = fs.calc_potential(dip, dip_loc)
            pot_sum += pot

        np.testing.assert_almost_equal(pot_MD, pot_sum)
        np.testing.assert_allclose(pot_MD, pot_sum, rtol=1E-4)

    def test_calc_potential_from_multi_dipoles01(self):
        """test comparison between multi-dipoles and single dipole approach"""
        neuron.h('forall delete_section()')
        soma = neuron.h.Section(name='soma')
        dend1 = neuron.h.Section(name='dend1')
        dend2 = neuron.h.Section(name='dend2')
        dend3 = neuron.h.Section(name='dend3')
        dend1.connect(soma(1.0), 0)
        dend2.connect(soma(1.0), 0)
        dend3.connect(soma(1.0), 0)
        morphology = neuron.h.SectionList()
        morphology.wholetree()
        radii = [300, 400, 500, 600]
        sigmas = [0.3, 1.5, 0.015, 0.3]
        electrode_locs = np.array([[0., 0., 290.],
                                   [10., 90., 300.],
                                   [-90, 50., 400.],
                                   [110.3, -100., 500.]])
        cell = cell_w_synapse_from_sections(morphology)
        cell.set_pos(x=0, y=0, z=100)
        t_point = [1,100,-1]

        MD_4s = LFPy.FourSphereVolumeConductor(radii, sigmas, electrode_locs)
        p, dipole_locs = cell.get_multi_current_dipole_moments(t_point)
        Np, Nt, Nd = p.shape
        Ne = electrode_locs.shape[0]
        pot_MD = MD_4s.calc_potential_from_multi_dipoles(cell, t_point)

        pot_sum = np.zeros((Ne, Nt))
        for i in range(Np):
            dip = p[i]
            dip_loc = dipole_locs[i]
            fs = LFPy.FourSphereVolumeConductor(radii, sigmas, electrode_locs)
            pot = fs.calc_potential(dip, dip_loc)
            pot_sum += pot

        np.testing.assert_almost_equal(pot_MD, pot_sum)
        np.testing.assert_allclose(pot_MD, pot_sum, rtol=1E-4)

class testInfiniteVolumeConductor(unittest.TestCase):
    """
    test class InfiniteVolumeConductor
    """
    def test_get_dipole_potential(self):
        sigma = 0.3
        r = np.array([[0., 0., 1.], [0., 1., 0.]])
        p = np.array([[0., 0., 4*np.pi*0.3], [0., 4*np.pi*0.3, 0.]])
        inf_model = LFPy.InfiniteVolumeConductor(sigma)
        phi = inf_model.get_dipole_potential(p, r)
        np.testing.assert_allclose(phi, np.array([[1., 0.], [0., 1.]]))

    def test_get_multi_dipole_potential00(self):
        neuron.h('forall delete_section()')
        soma = neuron.h.Section(name='soma')
        dend1 = neuron.h.Section(name='dend1')
        dend2 = neuron.h.Section(name='dend2')
        dend3 = neuron.h.Section(name='dend3')
        dend4 = neuron.h.Section(name='dend4')
        dend5 = neuron.h.Section(name='dend5')
        dend1.connect(soma(0.5), 0)
        dend2.connect(dend1(1.0), 0)
        dend3.connect(dend2(1.0), 0)
        dend4.connect(dend3(1.0), 0)
        dend5.connect(dend4(1.0), 0)
        morphology = neuron.h.SectionList()
        morphology.wholetree()
        electrode_locs = np.array([[0., 0., 10000.]])
        cell, electrode = cell_w_synapse_from_sections_w_electrode(morphology, electrode_locs)
        sigma = 0.3
        t_point = 0

        MD_inf = LFPy.InfiniteVolumeConductor(sigma)
        pot_MD = MD_inf.get_multi_dipole_potential(cell, electrode_locs)
        pot_cb = electrode.LFP

        np.testing.assert_almost_equal(pot_MD, pot_cb)
        np.testing.assert_allclose(pot_MD, pot_cb, rtol=1E-4)

    def test_get_multi_dipole_potential01(self):
        morphology = os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks.hoc')
        electrode_locs = np.array([[0., 0., 10000.]])
        cell, electrode = cell_w_synapse_from_sections_w_electrode(morphology, electrode_locs)
        sigma = 0.3
        t_point = 0

        MD_inf = LFPy.InfiniteVolumeConductor(sigma)
        pot_MD = MD_inf.get_multi_dipole_potential(cell, electrode_locs)
        pot_cb = electrode.LFP

        np.testing.assert_almost_equal(pot_MD, pot_cb)
        np.testing.assert_allclose(pot_MD, pot_cb, rtol=1E-3)

    def test_get_multi_dipole_potential02(self):
        morphology = os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks.hoc')
        electrode_locs = np.array([[0., 0., 10000.]])
        cell, electrode = cell_w_synapse_from_sections_w_electrode(morphology, electrode_locs)
        sigma = 0.3
        t_point = [10, 100, 1000]

        MD_inf = LFPy.InfiniteVolumeConductor(sigma)
        pot_MD = MD_inf.get_multi_dipole_potential(cell, electrode_locs, t_point)
        pot_cb = electrode.LFP[:,t_point]

        np.testing.assert_almost_equal(pot_MD, pot_cb)
        np.testing.assert_allclose(pot_MD, pot_cb, rtol=1E-3)


class testOneSphereVolumeConductor(unittest.TestCase):
    """
    test class OneSphereVolumeConductor
    """
    def test_OneSphereVolumeConductor_00(self):
        """test case where sigma_i == sigma_o which
        should be identical to the standard point-source potential in
        infinite homogeneous media
        """
        # current magnitude
        I = 1.
        # conductivity
        sigma = 0.3
        # sphere radius
        R = 1000
        # source location (along x-axis)
        rs = 800
        # sphere coordinates of observation points
        radius = np.r_[np.arange(0, rs), np.arange(rs+1, rs*2)]
        theta = np.zeros(radius.shape)
        phi = np.zeros(radius.shape)
        r = np.array([radius, theta, phi])

        # predict potential
        sphere = LFPy.OneSphereVolumeConductor(r=r, R=R, sigma_i=sigma, sigma_o=sigma)
        phi = sphere.calc_potential(rs=rs, I=I)

        # ground truth
        phi_gt = I / (4*np.pi*sigma*abs(radius-rs))

        # test
        np.testing.assert_almost_equal(phi, phi_gt)

    def test_OneSphereVolumeConductor_01(self):
        """test case where sigma_i == sigma_o which
        should be identical to the standard point-source potential in
        infinite homogeneous media
        """
        # current magnitude
        I = np.ones(10)
        # conductivity
        sigma = 0.3
        # sphere radius
        R = 1000
        # source location (along x-axis)
        rs = 800
        # sphere coordinates of observation points
        radius = np.r_[np.arange(0, rs), np.arange(rs+1, rs*2)]
        theta = np.zeros(radius.shape)
        phi = np.zeros(radius.shape)
        r = np.array([radius, theta, phi])

        # predict potential
        sphere = LFPy.OneSphereVolumeConductor(r=r, R=R, sigma_i=sigma, sigma_o=sigma)
        phi = sphere.calc_potential(rs=rs, I=I)

        # ground truth
        phi_gt = I[0] / (4*np.pi*sigma*abs(radius-rs))

        # test
        np.testing.assert_almost_equal(phi, np.array([phi_gt]*I.size).T)

    def test_OneSphereVolumeConductor_02(self):
        """test case where sigma_i == sigma_o which
        should be identical to the standard point-source potential in
        infinite homogeneous media
        """
        # current magnitude
        I = 1.
        # conductivity
        sigma = 0.3
        # sphere radius
        R = 10000
        # cell body position
        xs = 8000.
        # sphere coordinates of observation points
        radius = np.r_[np.arange(0, xs), np.arange(xs+1, xs*2)][::10]
        theta = np.zeros(radius.shape)+np.pi/2
        phi = np.zeros(radius.shape)
        r = np.array([radius, theta, phi])
        # set up cell
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test', 'stick.hoc'))
        cell.set_pos(x=xs, y=0, z=0)
        cell.set_rotation(y=np.pi/2)

        # predict potential
        sphere = LFPy.OneSphereVolumeConductor(r=r, R=R, sigma_i=sigma, sigma_o=sigma)
        mapping = sphere.calc_mapping(cell=cell, n_max=100)

        # ground truth and tests
        for i, x in enumerate(cell.xmid):
            dist = radius-x
            dist[abs(dist) < cell.diam[i]] = cell.diam[i]
            phi_gt = I / (4*np.pi*sigma*abs(dist))
            np.testing.assert_almost_equal(mapping[:, i], phi_gt)


######## Functions used by tests: ##############################################

def make_class_object(rz, r_el):
    '''Return class object fs'''
    radii = [79., 80., 85., 90.]
    sigmas = [0.3, 0.015, 15, 0.3]
    fs = LFPy.FourSphereVolumeConductor(radii, sigmas, r_el)
    fs._rz_params(rz)
    return fs

def make_simple_class_object():
    '''Return class object fs'''
    radii = [1., 2., 4., 8.]
    sigmas = [1., 2., 4., 8.]
    rz1 = np.array([0., 0., .9])
    r_el = np.array([[0., 0., 1.5]])
    fs = LFPy.FourSphereVolumeConductor(radii, sigmas, r_el)
    fs._rz_params(rz1)
    return fs

def decompose_dipole(P1):
    '''Return decomposed current dipole'''
    rz1 = np.array([0., 0., 70.])
    r_el = np.array([[0., 0., 90.]])
    fs = make_class_object(rz1, r_el)
    p_rad, p_tan = fs._decompose_dipole(P1)
    return p_rad, p_tan

def cell_w_synapse_from_sections(morphology):
    '''
    Make cell and synapse objects, set spike, simulate and return cell
    '''
    cellParams = {
        'morphology': morphology,
        'cm' : 1,
        'Ra' : 150,
        'v_init' : -65,
        'passive' : True,
        'passive_parameters' : {'g_pas' : 1./30000, 'e_pas' : -65},
        'dt' : 2**-6,
        'tstart' : -50,
        'tstop' : 50,
        'delete_sections' : False
    }

    synapse_parameters = {'e': 0.,
                      'syntype': 'ExpSyn',
                      'tau': 5.,
                      'weight': .001,
                      'record_current': True,
                      'idx': 1}

    cell = LFPy.Cell(**cellParams)
    synapse = LFPy.Synapse(cell, **synapse_parameters)
    synapse.set_spike_times(np.array([1.]))
    cell.simulate(rec_current_dipole_moment=True, rec_vmem=True)
    return cell

def cell_w_synapse_from_sections_w_electrode(morphology, electrode_locs):
    '''
    Make cell and synapse objects, set spike, simulate and return cell
    '''
    cellParams = {
        'morphology': morphology,
        'cm' : 1,
        'Ra' : 150,
        'v_init' : -65,
        'passive' : True,
        'passive_parameters' : {'g_pas' : 1./30000, 'e_pas' : -65},
        'dt' : 2**-6,
        'tstart' : -50,
        'tstop' : 50,
        'delete_sections' : False
    }

    electrodeParams = {'sigma': 0.3,
                        'x': electrode_locs[:,0],
                        'y': electrode_locs[:,1],
                        'z': electrode_locs[:,2],
                        }
    cell = LFPy.Cell(**cellParams)
    electrode = LFPy.RecExtElectrode(cell, **electrodeParams)

    synapse_parameters = {'e': 0.,
                      'syntype': 'ExpSyn',
                      'tau': 5.,
                      'weight': .1,
                      'record_current': True,
                      'idx': cell.totnsegs-1}

    synapse = LFPy.Synapse(cell, **synapse_parameters)
    synapse.set_spike_times(np.array([1.]))
    cell.simulate(electrode=[electrode],rec_current_dipole_moment=True, rec_vmem=True)
    return cell, electrode



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2017 Computational Neuroscience Group, NMBU.
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""


from scipy.special import eval_legendre, lpmv, legendre
import numpy as np
from warnings import warn


class OneSphereVolumeConductor(object):
    """
    Computes extracellular potentials within and outside a spherical volume-
    conductor model that assumes homogeneous, isotropic, linear (frequency
    independent) conductivity in and outside the sphere with a radius R. The
    conductivity in and outside the sphere must be greater than 0, and the
    current source(s) must be located within the radius R.

    The implementation is based on the description of electric potentials of
    point charge in an dielectric sphere embedded in dielectric media, which is
    mathematically equivalent to a current source in conductive media, as
    published by Deng (2008), Journal of Electrostatics 66:549-560

    Parameters
    ----------
    r : ndarray, dtype=float
        shape(3, n_points) observation points in space in spherical coordinates
        (radius, theta, phi) relative to the center of the sphere.
    R : float
        sphere radius (µm)
    sigma_i : float
        electric conductivity for radius r <= R (S/m)
    sigma_o : float
        electric conductivity for radius r > R (S/m)

    Examples
    --------
    Compute the potential for a single monopole along the x-axis:

    >>> # import modules
    >>> from LFPy import OneSphereVolumeConductor
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> # observation points in spherical coordinates (flattened)
    >>> X, Y = np.mgrid[-15000:15100:1000., -15000:15100:1000.]
    >>> r = np.array([np.sqrt(X**2 + Y**2).flatten(),
    >>>               np.arctan2(Y, X).flatten(),
    >>>               np.zeros(X.size)])
    >>> # set up class object and compute electric potential in all locations
    >>> sphere = OneSphereVolumeConductor(r, R=10000.,
    >>>                                   sigma_i=0.3, sigma_o=0.03)
    >>> Phi = sphere.calc_potential(rs=8000, I=1.).reshape(X.shape)
    >>> # plot
    >>> fig, ax = plt.subplots(1,1)
    >>> im=ax.contourf(X, Y, Phi,
    >>>              levels=np.linspace(Phi.min(), np.median(Phi[np.isfinite(Phi)])*4, 30))
    >>> circle = plt.Circle(xy=(0,0), radius=sphere.R, fc='none', ec='k')
    >>> ax.add_patch(circle)
    >>> fig.colorbar(im, ax=ax)
    >>> plt.show()
    """
    def __init__(self,
                 r,
                 R=10000.,
                 sigma_i=0.3,
                 sigma_o=0.03):
        """initialize class OneSphereVolumeConductor"""
        # check inputs
        try:
            assert(r.shape[0] == 3)
            assert(r.ndim == 2)
        except AssertionError as ae:
            raise AssertionError('r must be a shape (3, n_points) ndarray')
        try:
            assert((type(R) is float) or (type(R) is int))
        except AssertionError as ae:
            raise AssertionError('sphere radius R must be a float value')
        try:
            assert((sigma_i > 0) & (sigma_o > 0))
        except AssertionError as ae:
            raise AssertionError('sigma_i and sigma_o must both be positive values')

        self.r = r
        self.R = R
        self.sigma_i = sigma_i
        self.sigma_o = sigma_o


    def calc_potential(self, rs, I, min_distance=1., n_max=1000):
        """
        Return the electric potential at observation points for source current
        with magnitude I as function of time.

        Parameters
        ----------
        rs : float
            monopole source location along the horizontal x-axis (µm)
        I : float or ndarray, dtype float
            float or shape (n_tsteps, ) array containing source current (nA)
        min_distance : None or float
            minimum distance between source location and observation point (µm)
            (in order to avoid singular values)
        n_max : int
            Number of elements in polynomial expansion to sum over
            (see Deng 2008).

        Returns
        -------
        Phi : ndarray
            shape (n-points, ) ndarray of floats if I is float like. If I is
            an 1D ndarray, and shape (n-points, I.size) ndarray is returned.
            Unit (mV).
        """
        try:
            assert(type(rs) in [int, float, np.float64])
            assert(abs(rs) < self.R)
        except AssertionError as ae:
            raise AssertionError('source location rs must be a float value and |rs| must be less than sphere radius R')
        try:
            assert((min_distance is None) or (type(min_distance) in [float, int, np.float64]))
        except AssertionError:
            raise AssertionError('min_distance must be None or a float')

        r = self.r[0]
        theta = self.r[1]


        # add harmonical contributions due to inhomogeneous media
        inds_i = r <= self.R
        inds_o = r > self.R

        # observation points r <= R
        phi_i = np.zeros(r.size)
        for j, (theta_i, r_i) in enumerate(zip(theta[inds_i], r[inds_i])):
            coeffs_i = np.zeros(n_max)
            for n in range(n_max):
                coeffs_i[n] = ((self.sigma_i - self.sigma_o)*(n+1))  / (self.sigma_i*n + self.sigma_o*(n+1)) * ((r_i*rs)/self.R**2)**n
            poly_i = np.polynomial.legendre.Legendre(coeffs_i)
            phi_i[np.where(inds_i)[0][j]] = poly_i(np.cos(theta_i))
        phi_i[inds_i] *= 1./self.R

        # observation points r > R
        phi_o = np.zeros(r.size)
        for j, (theta_o, r_o) in enumerate(zip(theta[inds_o], r[inds_o])):
            coeffs_o = np.zeros(n_max)
            for n in range(n_max):
                coeffs_o[n] = (self.sigma_i*(2*n+1) ) / (self.sigma_i*n + self.sigma_o*(n+1)) * (rs / r_o)**n
            poly_o = np.polynomial.legendre.Legendre(coeffs_o)
            phi_o[np.where(inds_o)[0][j]] = poly_o(np.cos(theta_o))
        phi_o[inds_o] *= 1./r[inds_o]

        # potential in homogeneous media
        if min_distance is None:
            phi_i[inds_i] += 1. / np.sqrt(r[r <= self.R]**2 + rs**2 - 2*r[inds_i]*rs*np.cos(theta[inds_i]))
        else:
            denom = np.sqrt(r[inds_i]**2 + rs**2 - 2*r[inds_i]*rs*np.cos(theta[inds_i]))
            denom[denom < min_distance] = min_distance
            phi_i[inds_i] += 1./denom



        if type(I) is np.ndarray:
            try:
                assert(np.all(np.isfinite(I)))
                assert(np.all(np.isreal(I)))
                assert(I.ndim == 1)
            except AssertionError:
                raise AssertionError('input argument I must be float or 1D ndarray with float values')

            return np.dot((phi_i + phi_o).reshape((1, -1)).T,
                I.reshape((1, -1))) / (4.*np.pi*self.sigma_i)
        else:
            try:
                assert(np.isfinite(I)) and (np.shape(I) == ())
            except AssertionError:
                raise AssertionError('input argument I must be float or 1D ndarray with float values')
            return I / (4.*np.pi*self.sigma_i)*(phi_i + phi_o)


    def calc_mapping(self, cell, n_max=1000):
        """
        Compute linear mapping between transmembrane currents of LFPy.Cell like
        object instantiation and extracellular potential in and outside of
        sphere. Cell position must be set in space, using the method
        ``Cell.set_pos(**kwargs)``.

        Parameters
        ----------
        cell : LFPy.Cell like instance
            Instantiation of class LFPy.Cell, TemplateCell or NetworkCell.
        n_max : int
            Number of elements in polynomial expansion to sum over
            (see Deng 2008).

        Examples
        --------
        Compute extracellular potential in one-sphere volume conductor model
        from LFPy.Cell object:

        >>> # import modules
        >>> import LFPy
        >>> import os
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from matplotlib.collections import PolyCollection
        >>> # create cell
        >>> cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks.hoc'),
        >>>                  tstop=10.)
        >>> cell.set_pos(z=9800.)
        >>> # stimulus
        >>> syn = LFPy.Synapse(cell, idx=cell.totnsegs-1, syntype='Exp2Syn', weight=0.01)
        >>> syn.set_spike_times(np.array([1.]))
        >>> # simulate
        >>> cell.simulate(rec_imem=True)
        >>> # observation points in spherical coordinates (flattened)
        >>> X, Z = np.mgrid[-500:501:10., 9500:10501:10.]
        >>> Y = np.zeros(X.shape)
        >>> r = np.array([np.sqrt(X**2 + Z**2).flatten(),
        >>>               np.arccos(Z / np.sqrt(X**2 + Z**2)).flatten(),
        >>>               np.arctan2(Y, X).flatten()])
        >>> # set up class object and compute mapping between segment currents
        >>> # and electric potential in space
        >>> sphere = LFPy.OneSphereVolumeConductor(r=r, R=10000.,
        >>>                                        sigma_i=0.3, sigma_o=0.03)
        >>> mapping = sphere.calc_mapping(cell, n_max=1000)
        >>> # pick out some time index for the potential and compute potential
        >>> ind = cell.tvec==2.
        >>> Phi = np.dot(mapping, cell.imem)[:, ind].reshape(X.shape)
        >>> # plot potential
        >>> fig, ax = plt.subplots(1,1)
        >>> zips = []
        >>> for x, z in cell.get_idx_polygons(projection=('x', 'z')):
        >>>     zips.append(zip(x, z))
        >>> polycol = PolyCollection(zips,
        >>>                          edgecolors='none',
        >>>                          facecolors='gray')
        >>> vrange = 1E-3 # limits for color contour plot
        >>> im=ax.contour(X, Z, Phi,
        >>>              levels=np.linspace(-vrange, vrange, 41))
        >>> circle = plt.Circle(xy=(0,0), radius=sphere.R, fc='none', ec='k')
        >>> ax.add_collection(polycol)
        >>> ax.add_patch(circle)
        >>> ax.axis(ax.axis('equal'))
        >>> ax.set_xlim(X.min(), X.max())
        >>> ax.set_ylim(Z.min(), Z.max())
        >>> fig.colorbar(im, ax=ax)
        >>> plt.show()

        Returns
        -------
        ndarray
            Shape (n_points, n_compartments) mapping between individual
            segments and extracellular potential in extracellular locations

        Notes
        -----
        Each segment is treated as a point source in space. The minimum
        source to measurement site distance will be set to the diameter of
        each segment

        """
        # midpoint position of compartments in spherical coordinates
        radius = np.sqrt(cell.xmid**2 + cell.ymid**2 + cell.zmid**2)
        theta = np.arccos(cell.zmid/radius)
        phi = np.arctan2(cell.ymid, cell.xmid)
        diam = cell.diam

        # since the sources must be located on the x-axis, we keep a copy
        # of the unrotated coordinate system for the contact points:
        r_orig = np.copy(self.r)

        # unit current amplitude
        I = 1.

        # initialize mapping array
        mapping = np.zeros((self.r.shape[1], radius.size))

        # compute the mapping for each compartment
        for i, (radius_i, theta_i, _, diam_i) in enumerate(zip(radius, theta, phi, diam)):
            self.r = np.array([r_orig[0], # radius unchanged
                               r_orig[1] - theta_i, # rotate relative to source location
                               r_orig[2]]) # phi unchanged (don't enter equations)
            mapping[:, i] = self.calc_potential(radius_i, I=I, min_distance=diam_i, n_max=n_max)

        # reset measurement locations
        self.r = r_orig

        # return mapping between segment currents and contrib in each
        # measurement location
        return mapping


class FourSphereVolumeConductor(object):
    """
    Main class for computing extracellular potentials in a four-sphere
    volume conductor model that assumes homogeneous, isotropic, linear
    (frequency independent) conductivity within the inner sphere and outer
    shells. The conductance outside the outer shell is 0 (air).

    Parameters
    ----------
    radii : list, dtype=float
        Len 4 list with the outer radii in units of (µm) for the 4
        concentric shells in the four-sphere model: brain, csf, skull and
        scalp, respectively.
    sigmas : list, dtype=float
        Len 4 list with the electrical conductivity in units of (S/m) of
        the four shells in the four-sphere model: brain, csf, skull and
        scalp, respectively.
    r_electrodes : ndarray, dtype=float
        Shape (n_contacts, 3) array containing n_contacts electrode locations
        in cartesian coordinates in units of (µm).
        All ``r_el`` in ``r_electrodes`` must be less than or equal to scalp
        radius and larger than the distance between dipole and sphere
        center: ``|rz| < r_el <= radii[3]``.

    Examples
    --------
    Compute extracellular potential from current dipole moment in four-sphere
    head model:

    >>> import LFPy
    >>> import numpy as np
    >>> radii = [79000., 80000., 85000., 90000.]
    >>> sigmas = [0.3, 1.5, 0.015, 0.3]
    >>> r = np.array([[0., 0., 90000.], [0., 85000., 0.]])
    >>> rz = np.array([0., 0., 78000.])
    >>> sphere_model = LFPy.FourSphereVolumeConductor(radii, sigmas, r)
    >>> # current dipole moment
    >>> p = np.array([[10., 10., 10.]]*10) # 10 timesteps
    >>> # compute potential
    >>> potential = sphere_model.calc_potential(p, rz)

    """

    def __init__(self, radii, sigmas, r_electrodes, iter_factor = 2./99.*1e-6):
        """Initialize class FourSphereVolumeConductor"""
        self.r1 = float(radii[0])
        self.r2 = float(radii[1])
        self.r3 = float(radii[2])
        self.r4 = float(radii[3])

        self.sigma1 = float(sigmas[0])
        self.sigma2 = float(sigmas[1])
        self.sigma3 = float(sigmas[2])
        self.sigma4 = float(sigmas[3])

        self.r12 = self.r1 / self.r2
        self.r21 = self.r2 / self.r1
        self.r23 = self.r2 / self.r3
        self.r32 = self.r3 / self.r2
        self.r34 = self.r3 / self.r4
        self.r43 = self.r4 / self.r3

        self.sigma12 = self.sigma1 / self.sigma2
        self.sigma21 = self.sigma2 / self.sigma1
        self.sigma23 = self.sigma2 / self.sigma3
        self.sigma32 = self.sigma3 / self.sigma2
        self.sigma34 = self.sigma3 / self.sigma4
        self.sigma43 = self.sigma4 / self.sigma3

        self.rxyz = r_electrodes
        self.r = np.sqrt(np.sum(r_electrodes ** 2, axis=1))

        self.iteration_stop_factor = iter_factor
        self._check_params()

    def _check_params(self):
        """Check that radii, sigmas and r contain reasonable values"""
        if (self.r1 < 0) or (self.r1 > self.r2) or (self.r2 > self.r3) or (self.r3 > self.r4):
            raise RuntimeError('Radii of brain (radii[0]), CSF (radii[1]), '
                               'skull (radii[2]) and scalp (radii[3]) '
                               'must be positive ints or floats such that '
                               '0 < radii[0] < radii[1] < radii[2] < radii[3].')

        if (self.sigma1 < 0) or (self.sigma2 < 0) or (self.sigma3 < 0) or (self.sigma4 < 0):
            raise RuntimeError('Conductivities (sigmas), must contain positive'
                               ' ints or floats.')

        if any(r > self.r4 for r in self.r):
            raise ValueError('Electrode located outside head model.'
                               'r > radii[3]. r = %s, r4 = %s',
                               self.r, self.r4)

    def _rz_params(self, rz):
        """Define dipole location vector, and check that dipole is located in
        the brain, closer to the center than any measurement location."""
        self._rzloc = rz
        self._rz = np.sqrt(np.sum(rz ** 2))
        self._z = self._rzloc/self._rz
        if self._rz == 0:
            raise RuntimeError('Placing dipole in center of head model causes division by zero.')

        self._rz1 = self._rz / self.r1

        if self._rz1 >= 1:
            raise RuntimeError('Dipole should be placed inside brain, i.e. |rz| < |r1|')

        elif self._rz1 > 0.99999:
            warn('Dipole should be placed minimum ~1µm away from brain surface, '
                  'to avoid extremely slow convergence.')

        elif self._rz1 > 0.9999:
            warn('Computation time might be long due to slow convergence. '
                 'Can be avoided by placing dipole further away from brain surface.')

        if any(r < self._rz for r in self.r):
            raise RuntimeError('Electrode must be farther away from '
                               'brain center than dipole: r > rz.'
                               'r = %s, rz = %s', self.r, self._rz)

        # compute theta angle between rzloc and rxyz
        self._theta = self.calc_theta()


    def calc_potential(self, p, rz):
        """
        Return electric potential from current dipole moment p.

        Parameters
        ----------
        p : ndarray, dtype=float
            Shape (n_timesteps, 3) array containing the x,y,z components of the
            current dipole moment in units of (nA*µm) for all timesteps.
        rz : ndarray, dtype=float
            Shape (3, ) array containing the position of the current dipole in
            cartesian coordinates. Units of (µm).

        Returns
        -------
        potential : ndarray, dtype=float
            Shape (n_contacts, n_timesteps) array containing the electric
            potential at contact point(s) FourSphereVolumeConductor.r in units
            of (mV) for all timesteps of current dipole moment p.

        """

        self._rz_params(rz)
        n_contacts = self.r.shape[0]
        n_timesteps = p.shape[0]

        if np.linalg.norm(p) != 0:
            p_rad, p_tan = self._decompose_dipole(p)
        else:
            p_rad = np.zeros((n_timesteps, 3))
            p_tan = np.zeros((n_timesteps, 3))
        if np.linalg.norm(p_rad) != 0.:
            pot_rad = self._calc_rad_potential(p_rad)
        else:
            pot_rad = np.zeros((n_contacts, n_timesteps))

        if np.linalg.norm(p_tan) != 0.:
            pot_tan = self._calc_tan_potential(p_tan)
        else:
            pot_tan = np.zeros((n_contacts, n_timesteps))

        pot_tot = pot_rad + pot_tan
        return pot_tot

    def calc_potential_from_multi_dipoles(self, cell, timepoints=None):
        """
        Return electric potential from multiple current dipoles from cell.

        By multiple current dipoles we mean the dipoles computed from all
        axial currents in a neuron simulation, typically two
        axial currents per compartment, except for the root compartment.

        Parameters
        ----------
        cell : LFPy Cell object, LFPy.Cell
        timepoints : ndarray, dtype=int
            array of timepoints at which you want to compute
            the electric potential. Defaults to None. If not given,
            all simulation timesteps will be included.

        Returns
        -------
        potential : ndarray, dtype=float
            Shape (n_contacts, n_timesteps) array containing the electric
            potential at contact point(s) electrode_locs in units
            of (mV) for all timesteps of neuron simulation.

        Examples
        --------
        Compute extracellular potential from neuron simulation in
        four-sphere head model. Instead of simplifying the neural activity to
        a single dipole, we compute the contribution from every multi dipole
        from all axial currents in neuron simulation:

        >>> import LFPy
        >>> import numpy as np
        >>> cell = LFPy.Cell('PATH/TO/MORPHOLOGY', extracellular=False)
        >>> syn = LFPy.Synapse(cell, idx=cell.get_closest_idx(0,0,100),
        >>>                   syntype='ExpSyn', e=0., tau=1., weight=0.001)
        >>> syn.set_spike_times(np.mgrid[20:100:20])
        >>> cell.simulate(rec_vmem=True, rec_imem=False)
        >>> radii = [200., 300., 400., 500.]
        >>> sigmas = [0.3, 1.5, 0.015, 0.3]
        >>> electrode_locs = np.array([[50., -50., 250.]])
        >>> timepoints = np.array([0,100])
        >>> MD_4s = LFPy.FourSphereVolumeConductor(radii,
        >>>                                        sigmas,
        >>>                                        electrode_locs)
        >>> phi = MD_4s.calc_potential_from_multi_dipoles(cell,
        >>>                                               timepoints)
        """
        multi_p, multi_p_locs = cell.get_multi_current_dipole_moments(timepoints)
        N_elec = self.rxyz.shape[0]
        Ni, Nt, Nd = multi_p.shape
        potential = np.zeros((N_elec, Nt))
        for i in range(Ni):
            pot = self.calc_potential(multi_p[i], multi_p_locs[i])
            potential += pot
        return potential

    def _decompose_dipole(self, p):
        """
        Decompose current dipole moment vector in radial and tangential terms

        Parameters
        ----------
        p : ndarray, dtype=float
            Shape (n_timesteps, 3) array containing the x,y,z-components of the
            current dipole moment in units of (nA*µm) for all timesteps

        Returns:
        -------
        p_rad : ndarray, dtype=float
            Shape (n_timesteps, 3) array, radial part of p, parallel to self._rz
        p_tan : ndarray, dtype=float
            Shape (n_timesteps, 3) array, tangential part of p,
            orthogonal to self._rz
        """
        z_ = np.expand_dims(self._z, -1) # reshape z-axis vector
        p_rad = np.dot(np.dot(p, z_), z_.T)
        p_tan = p - p_rad
        return p_rad, p_tan

    def _calc_rad_potential(self, p_rad):
        """
        Return potential from radial dipole p_rad at location rz measured at r

        Parameters
        ----------
        p_rad : ndarray, dtype=float
            Shape (n_timesteps, 3) array, radial part of p
            in units of (nA*µm), parallel to self._rz

        Returns
        -------
        potential : ndarray, dtype=float
            Shape (n_contacts, n_timesteps) array containing the extracecllular
            potential at n_contacts contact point(s) FourSphereVolumeConductor.r
            in units of (mV) for all timesteps of p_rad
        """

        p_tot = np.linalg.norm(p_rad, axis=1)
        s_vector = self._sign_rad_dipole(p_rad)
        phi_const = s_vector * p_tot / (4 * np.pi * self.sigma1 * self._rz ** 2)
        n_terms = np.zeros((len(self.r), len(p_tot)))
        for el_point in range(len(self.r)):
            r_point = self.r[el_point]
            theta_point = self._theta[el_point]
            if r_point <= self.r1:
                n_terms[el_point] = self._potential_brain_rad(r_point,
                                                             theta_point)
            elif r_point <= self.r2:
                n_terms[el_point] = self._potential_csf_rad(r_point,
                                                           theta_point)
            elif r_point <= self.r3:
                n_terms[el_point] = self._potential_skull_rad(r_point,
                                                             theta_point)
            else:
                n_terms[el_point] = self._potential_scalp_rad(r_point,
                                                             theta_point)
        potential = phi_const * n_terms
        return potential

    def _calc_tan_potential(self, p_tan):
        """
        Return potential from tangential dipole P at location rz measured at r

        Parameters
        ----------
        p_tan : ndarray, dtype=float
            Shape (n_timesteps, 3) array, tangential part of p
            in units of (nA*µm), orthogonal to self._rz

        Returns
        _______
        potential : ndarray, dtype=float
            Shape (n_contacts, n_timesteps) array containing the extracecllular
            potential at n_contacts contact point(s) FourSphereVolumeConductor.r
            in units of (mV) for all timesteps of p_tan
        """
        phi = self.calc_phi(p_tan)
        p_tot = np.linalg.norm(p_tan, axis=1)
        phi_hom = - p_tot / (4 * np.pi * self.sigma1 * self._rz ** 2) * np.sin(phi)
        n_terms = np.zeros((len(self.r),1))
        for el_point in range(len(self.r)):
            r_point = self.r[el_point]
            theta_point = self._theta[el_point]
            # if r_electrode is orthogonal to p_tan, i.e. theta = 0 or theta = pi,
            # there is no contribution to electric potential from p_tan
            if (theta_point == 0.) or (theta_point == np.pi):
                n_terms[el_point] = 0
            elif r_point <= self.r1:
                n_terms[el_point] = self._potential_brain_tan(r_point, theta_point)
            elif r_point <= self.r2:
                n_terms[el_point] = self._potential_csf_tan(r_point, theta_point)
            elif r_point <= self.r3:
                n_terms[el_point] = self._potential_skull_tan(r_point, theta_point)
            else:
                n_terms[el_point] = self._potential_scalp_tan(r_point, theta_point)
        potential = n_terms * phi_hom

        return potential

    def calc_theta(self):
        """
        Return polar angle(s) between rzloc and contact point location(s)

        Returns
        -------
        theta : ndarray, dtype=float
            Shape (n_contacts, ) array containing polar angle
            in units of (radians) between z-axis and n_contacts contact
            point location vector(s) in FourSphereVolumeConductor.rxyz
            z-axis is defined in the direction of rzloc and the radial dipole.
        """
        cos_theta = np.dot(self.rxyz, self._rzloc) / (np.linalg.norm(self.rxyz, axis=1) * np.linalg.norm(self._rzloc))
        theta = np.arccos(cos_theta)
        return theta

    def calc_phi(self, p_tan):
        """
        Return azimuthal angle between x-axis and contact point locations(s)

        Parameters
        ----------
        p_tan : ndarray, dtype=float
            Shape (n_timesteps, 3) array containing
            tangential component of current dipole moment in units of (nA*µm)

        Returns
        -------
        phi : ndarray, dtype=float
            Shape (n_contacts, n_timesteps) array containing azimuthal angle
            in units of (radians) between x-axis vector(s) and projection of
            contact point location vector(s) rxyz into xy-plane.
            z-axis is defined in the direction of rzloc.
            y-axis is defined in the direction of p_tan (orthogonal to rzloc).
            x-axis is defined as cross product between p_tan and rzloc (x).
        """

        # project rxyz onto z-axis (rzloc)
        proj_rxyz_rz = self.rxyz*self._z
        # find projection of rxyz in xy-plane
        rxy = self.rxyz - proj_rxyz_rz
        # define x-axis
        x = np.cross(p_tan, self._z)

        phi = np.zeros((len(self.rxyz), len(p_tan)))
        # create masks to avoid computing phi when phi is not defined
        mask = np.ones(phi.shape, dtype=bool)
        # phi is not defined when theta= 0,pi or |p_tan| = 0
        mask[(self._theta == 0) | (self._theta == np.pi)] = np.zeros(len(p_tan))
        mask[:,np.abs(np.linalg.norm(p_tan, axis=1)) == 0] = 0

        cos_phi = np.zeros(phi.shape)
        # compute cos_phi using mask to avoid zerodivision
        cos_phi[mask] = np.dot(rxy, x.T)[mask] / np.outer(np.linalg.norm(rxy,
                  axis=1), np.linalg.norm(x, axis=1))[mask]

        # compute phi in [0, pi]
        phi[mask] = np.arccos(cos_phi[mask])

        # nb: phi in [-pi, pi]. since p_tan defines direction of y-axis,
        # phi < 0 when rxy*p_tan < 0
        phi[np.dot(rxy, p_tan.T) < 0] *= -1

        return phi

    def _sign_rad_dipole(self, p):
        """
        Determine whether radial dipoles are pointing inwards or outwards

        Parameters
        ----------
        p : ndarray, dtype=float
            Shape (n_timesteps, 3) array containing the current dipole moment
             in cartesian coordinates for all n_timesteps in units of (nA*µm)

        Returns
        -------
        sign_vector : ndarray
            Shape (n_timesteps, ) array containing +/-1 for all
            current dipole moments in p.
            If radial part of p[i] points outwards, sign_vector[i] = 1.
            If radial part of p[i] points inwards, sign_vector[i] = -1.

        """
        sign_vector = np.sign(np.dot(p, self._rzloc))
        return sign_vector

    def _potential_brain_rad(self, r, theta):
        """
        Return factor for calculation of potential in brain from rad. dipole

        Parameters
        ----------
        r : float
            Distance from origin to brain electrode location in units of (µm)
        theta : float
            Polar angle between brain electrode location and
            dipole location vector rzloc in units of (radians)

        Returns
        -------
        pot_sum : float
            Summationfactor for calculation of electrical potential in brain
            from radial current dipole moment. (unitless)
        """
        n = 1
        const = 1.
        coeff_sum = 0.
        consts = []
        # while const > self.iteration_stop_factor*coeff_sum:
        while const > 2./99.*1e-12*coeff_sum:
            c1n = self._calc_c1n(n)
            const = n*(c1n * (r / self.r1) ** n + (self._rz / r) ** (n + 1))
            coeff_sum += const
            consts.append(const)
            n += 1
        consts = np.insert(consts, 0, 0) # since the legendre function starts with P0
        leg_consts = np.polynomial.legendre.Legendre(consts)
        pot_sum = leg_consts(np.cos(theta))
        return pot_sum

    def _potential_csf_rad(self, r, theta):
        """
        Return factor for calculation of potential in CSF from rad. dipole

        Parameters
        ----------
        r : float
            Distance from origin to CSF electrode location in units of (µm)
        theta : float
            Polar angle between CSF electrode location and
            dipole location vector rzloc in units of (radians)

        Returns
        -------
        pot_sum : float
            Summation factor for calculation of electrical potential in CSF
            from radial current dipole moment. (unitless)
        """
        n = 1
        const = 1.
        coeff_sum = 0.
        consts = []
        # while const > self.iteration_stop_factor*coeff_sum:
        while const > 2./99.*1e-6*coeff_sum:
            term1 = self._calc_csf_term1(n,r)
            term2 = self._calc_csf_term2(n,r)
            const = n*(term1 + term2)
            coeff_sum += const
            consts.append(const)
            n += 1
        consts = np.insert(consts, 0, 0) # since the legendre function starts with P0
        leg_consts = np.polynomial.legendre.Legendre(consts)
        pot_sum = leg_consts(np.cos(theta))
        return pot_sum

    def _potential_skull_rad(self, r, theta):
        """
        Return factor for calculation of potential in skull from rad. dipole

        Parameters
        ----------
        r : float
            Distance from origin to skull electrode location in units of (µm)
        theta : float
            Polar angle between skull electrode location and
            dipole location vector rzloc in units of (radians)

        Returns
        -------
        pot_sum : float
            Summation factor for calculation of electrical potential in skull
            from radial current dipole moment. (unitless)
        """
        n = 1
        const = 1.
        coeff_sum = 0.
        consts = []
        # while const > self.iteration_stop_factor*coeff_sum:
        while const > 2./99.*1e-6*coeff_sum:
            c3n = self._calc_c3n(n)
            d3n = self._calc_d3n(n, c3n)
            const = n*(c3n * (r / self.r3) ** n + d3n * (self.r3 / r) ** (n + 1))
            coeff_sum += const
            consts.append(const)
            n += 1
        consts = np.insert(consts, 0, 0) # since the legendre function starts with P0
        leg_consts = np.polynomial.legendre.Legendre(consts)
        pot_sum = leg_consts(np.cos(theta))
        return pot_sum

    def _potential_scalp_rad(self, r, theta):
        """
        Return factor for calculation of potential in scalp from radial dipole

        Parameters
        ----------
        r : float
            Distance from origin to scalp electrode location in units of (µm)
        theta : float
            Polar angle between scalp electrode location and
            dipole location vector rzloc in units of (radians)

        Returns
        -------
        pot_sum : float
            Summation factor for calculation of electrical potential in scalp
            from radial current dipole moment. (unitless)
        """
        n = 1
        const = 1.
        coeff_sum = 0.
        consts = []
        # while const > self.iteration_stop_factor*coeff_sum:
        while const > 2./99.*1e-6*coeff_sum:
            c4n = self._calc_c4n(n)
            d4n = self._calc_d4n(n, c4n)
            const = n*(c4n * (r / self.r4) ** n + d4n * (self.r4 / r) ** (n + 1))
            coeff_sum += const
            consts.append(const)
            n += 1
        consts = np.insert(consts, 0, 0) # since the legendre function starts with P0
        leg_consts = np.polynomial.legendre.Legendre(consts)
        pot_sum = leg_consts(np.cos(theta))
        return pot_sum

    def _potential_brain_tan(self, r, theta):
        """
        Return factor for calculation of potential in brain from tan. dipole

        Parameters
        ----------
        r : float
            Distance from origin to brain electrode location in units of (µm)
        theta : float
            Polar angle between brain electrode location and
            dipole location vector rzloc in units of (radians)

        Returns
        -------
        pot_sum : float
            Summation factor for calculation of electrical potential in brain
            from tangential current dipole moment. (unitless)
        """
        n = 1
        const = 1.
        coeff_sum = 0.
        consts = []
        while const > self.iteration_stop_factor*coeff_sum:
            c1n = self._calc_c1n(n)
            const = (c1n * (r / self.r1) ** n + (self._rz / r) ** (n + 1))
            coeff_sum += const
            consts.append(const)
            n += 1
        pot_sum = np.sum([c*lpmv(1, i, np.cos(theta)) for c,i in zip(consts,np.arange(1,n))])
        return pot_sum

    def _potential_csf_tan(self, r, theta):
        """
        Return factor for calculation of potential in CSF from tan. dipole

        Parameters
        ----------
        r : float
            Distance from origin to CSF electrode location in units of (µm)
        theta : float
            Polar angle between CSF electrode location and
            dipole location vector rzloc in units of (radians)

        Returns
        -------
        pot_sum : float
            Summation factor for calculation of electrical potential in CSF
            from tangential current dipole moment. (unitless)
        """
        n = 1
        const = 1.
        coeff_sum = 0.
        consts = []
        while const > self.iteration_stop_factor*coeff_sum:
            term1 = self._calc_csf_term1(n,r)
            term2 = self._calc_csf_term2(n,r)
            const = term1 + term2
            coeff_sum += const
            consts.append(const)
            n += 1
        pot_sum = np.sum([c*lpmv(1, i, np.cos(theta)) for c,i in zip(consts,np.arange(1,n))])
        return pot_sum

    def _potential_skull_tan(self, r, theta):
        """
        Return factor for calculation of potential in skull from tan. dipole

        Parameters
        ----------
        r : float
            Distance from origin to skull electrode location in units of (µm)
        theta : float
            Polar angle between skull electrode location and
            dipole location vector rzloc in units of (radians)

        Returns
        -------
        pot_sum : float
            Summation factor for calculation of electrical potential in skull
            from tangential current dipole moment. (unitless)
        """
        n = 1
        const = 1.
        coeff_sum = 0.
        consts = []
        while const > self.iteration_stop_factor*coeff_sum:
            c3n = self._calc_c3n(n)
            d3n = self._calc_d3n(n, c3n)
            const = c3n * (r / self.r3) ** n + d3n * (self.r3 / r) ** (n + 1)
            coeff_sum += const
            consts.append(const)
            n += 1
        pot_sum = np.sum([c*lpmv(1, i, np.cos(theta)) for c,i in zip(consts,np.arange(1,n))])
        return pot_sum

    def _potential_scalp_tan(self, r, theta):
        """
        Return factor for calculation of potential in scalp from tan. dipole

        Parameters
        ----------
        r : float
            Distance from origin to scalp electrode location in units of (µm)
        theta : float
            Polar angle between scalp electrode location and
            dipole location vector rzloc in units of (radians)

        Returns
        -------
        pot_sum : float
            Summation factor for calculation of electrical potential in scalp
            from tangential current dipole moment. (unitless)
        """
        n = 1
        const = 1.
        coeff_sum = 0.
        consts = []
        while const > self.iteration_stop_factor*coeff_sum:
            c4n = self._calc_c4n(n)
            d4n = self._calc_d4n(n, c4n)
            const = c4n * (r / self.r4) ** n + d4n * (self.r4 / r) ** (n + 1)
            coeff_sum += const
            consts.append(const)
            n += 1
        pot_sum = np.sum([c*lpmv(1, i, np.cos(theta)) for c,i in zip(consts,np.arange(1,n))])
        return pot_sum

    def _calc_vn(self, n):
        r_const = ((self.r34 ** (2*n + 1) - 1) /
                  ((n + 1) / n * self.r34 ** (2*n + 1) + 1))
        if self.sigma23 + r_const == 0.0:
            v = 1e12
        else:
            v = (n / (n + 1) * self.sigma34 - r_const) / (self.sigma34 + r_const)
        return v

    def _calc_yn(self, n):
        vn = self._calc_vn(n)
        r_const = ((n / (n + 1) * self.r23 ** (2*n + 1) - vn) /
                  (self.r23 ** (2*n + 1) + vn))
        if self.sigma23 + r_const == 0.0:
            y = 1e12
        else:
            y = (n / (n + 1) * self.sigma23 - r_const) / (self.sigma23 + r_const)
        return y

    def _calc_zn(self, n):
        yn = self._calc_yn(n)
        z = (self.r12 ** (2*n+1) - (n + 1) / n * yn) / (self.r12 ** (2*n+1) + yn)
        return z

    def _calc_c1n(self, n):
        zn = self._calc_zn(n)
        c1 = (((n + 1) / n * self.sigma12 + zn) / (self.sigma12 - zn) * self._rz1**(n+1))
        return c1

    def _calc_c2n(self, n):
        yn = self._calc_yn(n)
        c1 = self._calc_c1n(n)
        c2 = ((c1 + self._rz1**(n+1)) * self.r12 ** (n + 1) /
             (self.r12 ** (2 * n + 1) + yn))
        return c2

    def _calc_d2n(self, n, c2):
        yn = self._calc_yn(n)
        d2 = yn * c2
        return d2

    def _calc_c3n(self, n):
        vn = self._calc_vn(n)
        c2 = self._calc_c2n(n)
        d2 = self._calc_d2n(n, c2)
        c3 = (c2 + d2) * self.r23 ** (n + 1) / (self.r23 ** (2*n + 1) + vn)
        return c3

    def _calc_d3n(self, n, c3):
        vn = self._calc_vn(n)
        d3 = vn * c3
        return d3

    def _calc_c4n(self, n):
        c3 = self._calc_c3n(n)
        d3 = self._calc_d3n(n, c3)
        c4 = ((n + 1) / n * self.r34 ** (n + 1) * (c3 + d3) /
             ((n + 1) / n * self.r34 ** (2*n + 1) + 1))
        return c4

    def _calc_d4n(self, n, c4):
        d4 = n / (n + 1) * c4
        return d4

    def _calc_csf_term1(self, n, r):
        yn = self._calc_yn(n)
        c1 = self._calc_c1n(n)
        term1 = ((c1 + self._rz1 ** (n + 1)) * self.r12*((self.r1*r)/
                (self.r2 ** 2)) **n / (self.r12**(2*n+1) + yn))
        return term1

    def _calc_csf_term2(self, n, r):
        yn = self._calc_yn(n)
        c1 = self._calc_c1n(n)
        term2 = (yn*(c1 + self._rz1 ** (n + 1))/
                (r/self.r2*((self.r1 * r) / self.r2**2) ** n +
                (r / self.r1) ** (n+1)*yn))
        return term2


class InfiniteVolumeConductor(object):
    """
    Main class for computing extracellular potentials with current dipole
    approximation in an infinite 3D volume conductor model that assumes
    homogeneous, isotropic, linear (frequency independent) conductivity

    Parameters
    ----------
    sigma : float
        Electrical conductivity in extracellular space in units of (S/cm)

    Examples
    --------
    Computing the potential from dipole moment valid in the far field limit.
    Theta correspond to the dipole alignment angle from the vertical z-axis:

    >>> import LFPy
    >>> import numpy as np
    >>> inf_model = LFPy.InfiniteVolumeConductor(sigma=0.3)
    >>> p = np.array([[10., 10., 10.]])
    >>> r = np.array([[1000., 0., 5000.]])
    >>> phi_p = inf_model.get_dipole_potential(p, r)

    """

    def __init__(self, sigma=0.3):
        "Initialize class InfiniteVolumeConductor"
        self.sigma = sigma

    def get_dipole_potential(self, p, r):
        """
        Return electric potential from current dipole with current dipole
        approximation

        p : ndarray, dtype=float
            Shape (n_timesteps, 3) array containing the x,y,z components of the
            current dipole moment in units of (nA*µm) for all timesteps
        r : ndarray, dtype=float
            Shape (n_contacts, 3) array contaning the displacement vectors
            from dipole location to measurement location

        Returns
        -------
        potential : ndarray, dtype=float
            Shape (n_contacts, n_timesteps) array containing the electric
            potential at contact point(s) FourSphereVolumeConductor.r in units
            of (mV) for all timesteps of current dipole moment p

        """
        dotprod = np.dot(r, p.T)
        r_factor = np.linalg.norm(r, axis=1)**3
        phi = 1./(4*np.pi*self.sigma)*(dotprod.T/ r_factor).T
        return phi

    def get_multi_dipole_potential(self, cell, electrode_locs, timepoints=None):
        """
        Return electric potential from multiple current dipoles from cell

        By multiple current dipoles we mean the dipoles computed from all
        axial currents in a neuron simulation, typically two
        axial currents per compartment, except for the root compartment.

        Parameters
        ----------
        cell : Cell object from LFPy
        electrode_locs : ndarray, dtype=float
            Shape (n_contacts, 3) array containing n_contacts electrode
            locations in cartesian coordinates in units of (µm).
            All ``r_el`` in electrode_locs must be placed so that ``|r_el|`` is
            less than or equal to scalp radius and larger than
            the distance between dipole and sphere
            center: ``|rz| < |r_el| <= radii[3]``.
        timepoints : ndarray, dtype=int
            array of timepoints at which you want to compute
            the electric potential. Defaults to None. If not given,
            all simulation timesteps will be included.

        Returns
        -------
        potential : ndarray, dtype=float
            Shape (n_contacts, n_timesteps) array containing the electric
            potential at contact point(s) electrode_locs in units
            of (mV) for all timesteps of neuron simulation

        Examples
        --------
        Compute extracellular potential from neuron simulation in
        four-sphere head model. Instead of simplifying the neural activity to
        a single dipole, we compute the contribution from every multi dipole
        from all axial currents in neuron simulation:

        >>> import LFPy
        >>> import numpy as np
        >>> cell = LFPy.Cell('PATH/TO/MORPHOLOGY', extracellular=False)
        >>> syn = LFPy.Synapse(cell, idx=cell.get_closest_idx(0,0,100),
        >>>                   syntype='ExpSyn', e=0., tau=1., weight=0.001)
        >>> syn.set_spike_times(np.mgrid[20:100:20])
        >>> cell.simulate(rec_vmem=True, rec_imem=False)
        >>> sigma = 0.3
        >>> timepoints = np.array([10, 20, 50, 100])
        >>> electrode_locs = np.array([[50., -50., 250.]])
        >>> MD_INF = LFPy.InfiniteVolumeConductor(sigma)
        >>> phi = MD_INF.get_multi_dipole_potential(cell, electrode_locs,
        >>>                                         timepoints = timepoints)
        """

        multi_p, multi_p_locs = cell.get_multi_current_dipole_moments(timepoints=timepoints)
        N_elec = electrode_locs.shape[0]
        Ni, Nt, Nd = multi_p.shape
        potentials = np.zeros((N_elec, Nt))
        for i in range(Ni):
            p = multi_p[i]
            r = electrode_locs - multi_p_locs[i]
            pot = self.get_dipole_potential(p, r)
            potentials += pot
        return potentials

def get_current_dipole_moment(dist, current):
    """
    Return current dipole moment vector P and P_tot of cell.

    Parameters
    ----------
    current : ndarray, dtype=float
        Either an array containing all transmembrane currents
        from all compartments of the cell, or an array of all
        axial currents between compartments in cell in units of nA
    dist : ndarray, dtype=float
        When input current is an array of axial currents,
        dist is the length of each axial current.
        When current is an array of transmembrane
        currents, dist is the position vector of each
        compartment middle. Unit is (µm).

    Returns
    -------
    P : ndarray, dtype=float
        Array containing the current dipole moment for all
        timesteps in the x-, y- and z-direction in units of (nA*µm)
    P_tot : ndarray, dtype=float
        Array containing the magnitude of the
        current dipole moment vector for all timesteps in units of (nA*µm)

    Examples
    --------
    Get current dipole moment vector and scalar moment from axial currents
    computed from membrane potentials:

    >>> import LFPy
    >>> import numpy as np
    >>> cell = LFPy.Cell('PATH/TO/MORPHOLOGY', extracellular=False)
    >>> syn = LFPy.Synapse(cell, idx=cell.get_closest_idx(0,0,1000),
    >>>                   syntype='ExpSyn', e=0., tau=1., weight=0.001)
    >>> syn.set_spike_times(np.mgrid[20:100:20])
    >>> cell.simulate(rec_vmem=True, rec_imem=False)
    >>> d_list, i_axial = cell.get_axial_currents()
    >>> P_ax, P_ax_tot = LFPy.get_current_dipole_moment(d_list, i_axial)

    Get current dipole moment vector and scalar moment from transmembrane
    currents using the extracellular mechanism in NEURON:

    >>> import LFPy
    >>> import numpy as np
    >>> cell = LFPy.Cell('PATH/TO/MORPHOLOGY', extracellular=True)
    >>> syn = LFPy.Synapse(cell, idx=cell.get_closest_idx(0,0,1000),
    >>>                   syntype='ExpSyn', e=0., tau=1., weight=0.001)
    >>> syn.set_spike_times(np.mgrid[20:100:20])
    >>> cell.simulate(rec_vmem=False, rec_imem=True)
    >>> P_imem, P_imem_tot = LFPy.get_current_dipole_moment(np.c_[cell.xmid,
    >>>                                                          cell.ymid,
    >>>                                                          cell.zmid],
    >>>                                                    cell.imem)

    """
    P = np.dot(current.T, dist)
    P_tot = np.sqrt(np.sum(P**2, axis=1))
    return P, P_tot

class MEG(object):
    """
    Basic class for computing magnetic field from current dipole moment.
    For this purpose we use the Biot-Savart law derived from Maxwell's equations
    under the assumption of negligible magnetic induction effects (Nunez and
    Srinivasan, Oxford University Press, 2006):

    .. math:: \\mathbf{H} = \\frac{\\mathbf{p} \\times \\mathbf{R}}{4 \\pi R^3}

    where :math:`\\mathbf{p}` is the current dipole moment, :math:`\\mathbf{R}`
    the vector between dipole source location and measurement location, and
    :math:`R=|\\mathbf{R}|`

    Note that the magnetic field :math:`\\mathbf{H}` is related to the magnetic
    field :math:`\\mathbf{B}` as :math:`\\mu_0 \\mathbf{H} = \\mathbf{B}-\\mathbf{M}`
    where :math:`\\mu_0` is the permeability of free space (very close to
    permebility of biological tissues). :math:`\\mathbf{M}` denotes material
    magnetization (also ignored)


    Parameters
    ----------
    sensor_locations : ndarray, dtype=float
        shape (n_locations x 3) array with x,y,z-locations of measurement
        devices where magnetic field of current dipole moments is calculated.
        In unit of (µm)
    mu : float
        Permeability. Default is permeability of vacuum (:math:`\\mu_0 = 4*\\pi*10^{-7}` T*m/A)


    Examples
    --------
    Define cell object, create synapse, compute current dipole moment:

    >>> import LFPy, os, numpy as np, matplotlib.pyplot as plt
    >>> cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks.hoc'),
    >>>                  passive=True)
    >>> cell.set_pos(0., 0., 0.)
    >>> syn = LFPy.Synapse(cell, idx=0, syntype='ExpSyn', weight=0.01, record_current=True)
    >>> syn.set_spike_times_w_netstim()
    >>> cell.simulate(rec_current_dipole_moment=True)

    Compute the dipole location as an average of segment locations weighted by membrane area:

    >>> dipole_location = (cell.area * np.c_[cell.xmid, cell.ymid, cell.zmid].T / cell.area.sum()).sum(axis=1)

    Instantiate the LFPy.MEG object, compute and plot the magnetic signal in a sensor location:

    >>> sensor_locations = np.array([[1E4, 0, 0]])
    >>> meg = LFPy.MEG(sensor_locations)
    >>> H = meg.calculate_H(cell.current_dipole_moment, dipole_location)
    >>> plt.subplot(311)
    >>> plt.plot(cell.tvec, cell.somav)
    >>> plt.subplot(312)
    >>> plt.plot(cell.tvec, syn.i)
    >>> plt.subplot(313)
    >>> plt.plot(cell.tvec, H[0])

    Raises
    ------
    AssertionError
        If dimensionality of sensor_locations is wrong
    """
    def __init__(self, sensor_locations, mu=4*np.pi*1E-7):
        """
        Initialize class MEG
        """
        try:
            assert(sensor_locations.ndim == 2)
        except AssertionError:
            raise AssertionError('sensor_locations.ndim != 2')
        try:
            assert(sensor_locations.shape[1] == 3)
        except AssertionError:
            raise AssertionError('sensor_locations.shape[1] != 3')

        # set attributes
        self.sensor_locations = sensor_locations
        self.mu = mu


    def calculate_H(self, current_dipole_moment, dipole_location):
        """
        Compute magnetic field H from single current-dipole moment localized
        somewhere in space

        Parameters
        ----------
        current_dipole_moment : ndarray, dtype=float
            shape (n_timesteps x 3) array with x,y,z-components of current-
            dipole moment time series data in units of (nA µm)
        dipole_location : ndarray, dtype=float
            shape (3, ) array with x,y,z-location of dipole in units of (µm)

        Returns
        -------
        ndarray, dtype=float
            shape (n_locations x n_timesteps x 3) array with x,y,z-components
            of the magnetic field :math:`\\mathbf{H}` in units of (nA/µm)

        Raises
        ------
        AssertionError
            If dimensionality of current_dipole_moment and/or dipole_location
            is wrong
        """
        try:
            assert(current_dipole_moment.ndim == 2)
        except AssertionError:
            raise AssertionError('current_dipole_moment.ndim != 2')
        try:
            assert(current_dipole_moment.shape[1] == 3)
        except AssertionError:
            raise AssertionError('current_dipole_moment.shape[1] != 3')
        try:
            assert(dipole_location.shape == (3, ))
        except AssertionError:
            raise AssertionError('dipole_location.shape != (3, )')


        # container
        H = np.empty((self.sensor_locations.shape[0],
                      current_dipole_moment.shape[0],
                      3))
        # iterate over sensor locations
        for i, r in enumerate(self.sensor_locations):
            R = r - dipole_location
            assert(R.ndim==1 and R.size == 3)
            try:
                assert(np.allclose(R, np.zeros(3)) == False)
            except AssertionError:
                raise AssertionError('Identical dipole and sensor location.')
            H[i, ] = np.cross(current_dipole_moment,
                              R) / (4 * np.pi * np.sqrt((R**2).sum())**3)

        return H


    def calculate_H_from_iaxial(self, cell):
        """
        Computes the magnetic field in space from axial currents computed from
        membrane potential values and axial resistances of multicompartment
        cells.

        See:
        Blagoev et al. (2007) Modelling the magnetic signature of neuronal
        tissue. NeuroImage 37 (2007) 137–148
        DOI: 10.1016/j.neuroimage.2007.04.033

        for details on the biophysics governing magnetic fields from axial
        currents.

        Parameters
        ----------
        cell : object
            LFPy.Cell-like object. Must have attribute vmem containing recorded
            membrane potentials in units of mV

        Examples
        --------
        Define cell object, create synapse, compute current dipole moment:

        >>> import LFPy, os, numpy as np, matplotlib.pyplot as plt
        >>> cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks.hoc'),
        >>>                  passive=True)
        >>> cell.set_pos(0., 0., 0.)
        >>> syn = LFPy.Synapse(cell, idx=0, syntype='ExpSyn', weight=0.01, record_current=True)
        >>> syn.set_spike_times_w_netstim()
        >>> cell.simulate(rec_vmem=True)

        Instantiate the LFPy.MEG object, compute and plot the magnetic signal in a sensor location:

        >>> sensor_locations = np.array([[1E4, 0, 0]])
        >>> meg = LFPy.MEG(sensor_locations)
        >>> H = meg.calculate_H_from_iaxial(cell)
        >>> plt.subplot(311)
        >>> plt.plot(cell.tvec, cell.somav)
        >>> plt.subplot(312)
        >>> plt.plot(cell.tvec, syn.i)
        >>> plt.subplot(313)
        >>> plt.plot(cell.tvec, H[0])

        Returns
        -------
        H : ndarray, dtype=float
            shape (n_locations x n_timesteps x 3) array with x,y,z-components
            of the magnetic field :math:`\\mathbf{H}` in units of (nA/µm)
        """
        i_axial, d_vectors, pos_vectors = cell.get_axial_currents_from_vmem()
        R = self.sensor_locations
        H = np.zeros((R.shape[0], cell.tvec.size, 3))

        for i, R_ in enumerate(R):
            for i_, d_, r_ in zip(i_axial, d_vectors, pos_vectors):
                r_rel = R_ - r_
                H[i, :, :] += np.dot(i_.reshape((-1, 1)),
                                     np.cross(d_, r_rel).reshape((1, -1))
                                     ) / (4*np.pi*np.sqrt((r_rel**2).sum())**3)
        return H



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This is an example scripts using LFPy with a passive cell model adapted from
Mainen and Sejnowski, Nature 1996, for the original files, see
http://senselab.med.yale.edu/modeldb/ShowModel.asp?model=2488

Illustration of simultaneous Electrocorticography (ECoG) and LFP measurements.
The same cell and input is used as in example_7.py
The LFP is calculated either assuming an infinite homogeneous medium,
or with taking into account the boundary between the cortex and the
medium that is covers the cortex. In addition the ECoG signal is
calculated at the top of the brain, with a large recording electrode,
which introduces strong spatial averaging.

Execution:

    python example_ECoG.py
    
Copyright (C) 2017 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

# importing some modules, setting some matplotlib values for pl.plot.
import LFPy
import numpy as np
import scipy.stats
import matplotlib.pyplot as plt
plt.rcParams.update({'font.size' : 12,
                     'figure.facecolor' : '1',
                     'figure.subplot.wspace' : 0.5,
                     'figure.subplot.hspace' : 0.5})

#seed for random generation
np.random.seed(1234)

################################################################################
# A couple of function declarations
################################################################################

def plot_LFP_and_ECoG(cell, electrode, electrode_MoI, ecog_electrode):

    ecog_clr = "darkgreen"
    cortex_clr = "salmon"
    surface_clr = "lightblue"

    fig = plt.figure(figsize=[16, 8])
    elec_clr = lambda idx: plt.cm.rainbow(1.0 * idx / (len(electrode.x)))

    xlim = np.max([ecog_electrode.r, np.max(np.abs(cell.xend))]) + 50

    #plot the morphology, electrode contacts, synapses cortical surface etc.
    ax = fig.add_axes([0.01, 0.08, 0.3, 0.9],
                      frameon=False, aspect=1, xlim=[-xlim, xlim],
                      ylim=[cell.somapos[2] - 250, top_of_cortex + 200])

    for sec in LFPy.cell.neuron.h.allsec():
        idx = cell.get_idx(sec.name())
        ax.plot(np.r_[cell.xstart[idx], cell.xend[idx][-1]],
                np.r_[cell.zstart[idx], cell.zend[idx][-1]],
                color='k')
    for i in range(len(cell.synapses)):
        l, = ax.plot([cell.synapses[i].x], [cell.synapses[i].z], '.', c='r')
    for i in range(electrode.x.size):
        l2, = ax.plot(electrode.x[i], electrode.z[i], 'o',
                color=elec_clr(i), ms=15, clip_on=False)

    # Plot ECoG electrode
    from matplotlib.patches import Ellipse
    el = Ellipse((ecog_electrode.x[0], ecog_electrode.z[0]),
                 2*ecog_electrode.r, 50, edgecolor='k',
                 facecolor=ecog_clr, clip_on=False)
    ax.add_artist(el)
    ax.text(-ecog_electrode.r, ecog_electrode.z[0] + 50,
            "ECoG electrode (R = {} $\mu$m)".format(ecog_electrode.r),
            color=ecog_clr)

    ax.text(xlim, top_of_cortex + 100,
            "$\sigma_S$=\n{} S/m".format(electrode_MoI.sigma_S),
            rotation=-90, color=surface_clr, va='center')

    ax.text(xlim, top_of_cortex - 150,
            "$\sigma_T$=\n{} S/m".format(electrode_MoI.sigma_T),
            rotation=-90, color=cortex_clr, va="center")

    ax.axhspan(top_of_cortex, top_of_cortex + 200,
               clip_on=False, facecolor=surface_clr, zorder=-2)

    ax.axhspan(np.min(cell.zend) - 200, top_of_cortex,
               clip_on=True, facecolor=cortex_clr, zorder=-2)

    ax.plot([-xlim, xlim+50], [top_of_cortex, top_of_cortex],
            clip_on=False, ls="--", c='gray', zorder=-1)
    ax.text(xlim+70, top_of_cortex, "Cortical\nsurface",
            color="gray", va="center")

    ax.legend([l, l2, el], ["Synapse", "LFP electrode", "ECoG electrode"],
              frameon=False, bbox_to_anchor=[0.5, -0.005])

    ax.set_xticks([])
    ax.set_yticks([])

    #plot the somatic trace
    ax = fig.add_axes([0.72, 0.22, 0.25, 0.22], title="Somatic membrane\npotential")
    ax.plot(cell.tvec, cell.somav, 'k')
    ax.set_xlabel('Time (ms)')
    ax.set_ylabel('mV')

    for loc, spine in ax.spines.items():
        if loc in ['right', 'top']:
            spine.set_color('none')
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

    #plot the laminar LFP
    ax = fig.add_axes([0.4, 0.08, 0.25, 0.75], frameon=False, yticks=[],
                      ylim=[cell.somapos[2] - 250, np.max(cell.zend) + 50])

    dz = electrode.z[1] - electrode.z[0]
    for i in range(electrode.x.size):

        ax.plot(cell.tvec, electrode.LFP[i] * 0.5 * dz / np.max(np.abs(electrode.LFP[i])) + electrode.z[i],
                c="gray", lw=2)
        ax.plot(cell.tvec, electrode_MoI.LFP[i] * 0.5 * dz / np.max(np.abs(electrode.LFP[i])) + electrode.z[i],
                c='k', lw=2, ls="--")
        ax.plot(-10, electrode.z[i],
                color=elec_clr(i), marker='o', ms=15, clip_on=False)

    ax.axis(ax.axis('tight'))
    ax.set_xlabel('Time (ms)')
    ax.set_ylabel('Normalized\nLaminar LFP')

    #plot the ECoG
    ax = fig.add_axes([0.4, 0.9, 0.25, 0.07], frameon=False, yticks=[])

    for i in range(ecog_electrode.x.size):
        ax.plot(cell.tvec, ecog_electrode.LFP[i] / np.max(np.abs(ecog_electrode.LFP[i])),
                c=ecog_clr, lw=2)

    ax.axis(ax.axis('tight'))
    ax.set_xlabel('Time (ms)')
    ax.set_ylabel('Normalized\nECoG')

    # Compare top electode LFP and ECoG
    ax = fig.add_axes([0.72, 0.6, 0.25, 0.3],
                      title="Comparison of ECoG and\ntop electrode LFP")

    l1, = ax.plot(cell.tvec, 1000 * electrode.LFP[-1], c="gray", lw=2)
    l3, = ax.plot(cell.tvec, 1000 * ecog_electrode.LFP[-1], c=ecog_clr, ls="-", lw=2)
    l2, = ax.plot(cell.tvec, 1000 * electrode_MoI.LFP[-1], c="k", ls="--", lw=2)

    fig.legend([l1, l2, l3], ["LFP (infinite homogeneous)",
                              "LFP (cortical surface included)", "ECoG"],
               frameon=False, loc="lower right")
    ax.axis(ax.axis('tight'))
    ax.set_xlabel('Time (ms)')
    ax.set_ylabel('$\mu$V')
    for loc, spine in ax.spines.items():
        if loc in ['right', 'top']:
            spine.set_color('none')
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

    return fig


def insert_synapses(synparams, section, n, spTimesFun, args):
    '''find n compartments to insert synapses onto'''
    idx = cell.get_rand_idx_area_norm(section=section, nidx=n)

    #Insert synapses in an iterative fashion
    for i in idx:
        synparams.update({'idx' : int(i)})

        # Some input spike train using the function call
        [spiketimes] = spTimesFun(**args)
        
        # Create synapse(s) and setting times using the Synapse class in LFPy
        s = LFPy.Synapse(cell, **synparams)
        s.set_spike_times(spiketimes)

################################################################################
# Define parameters, using dictionaries
# It is possible to set a few more parameters for each class or functions, but
# we chose to show only the most important ones here.
################################################################################

#define cell parameters used as input to cell-class
cellParameters = {
    'morphology' : 'morphologies/L5_Mainen96_wAxon_LFPy.hoc',
    'cm' : 1.0,                 # membrane capacitance
    'Ra' : 150,                 # axial resistance
    'v_init' : -65,             # initial crossmembrane potential
    'passive' : True,           # switch on passive mechs
    'passive_parameters' : {'g_pas' : 1./30000, 'e_pas' : -65}, # passive params
    'nsegs_method' : 'lambda_f',# method for setting number of segments,
    'lambda_f' : 100,           # segments are isopotential at this frequency
    'dt' : 2**-4,               # dt of LFP and NEURON simulation.
    'tstart' : -100,          #start time, recorders start at t=0
    'tstop' : 200,            #stop time of simulation
    #'custom_code'  : ['active_declarations_example3.hoc'], # will run this file
}
#Initialize cell instance, using the LFPy.Cell class
cell = LFPy.Cell(**cellParameters)

top_of_cortex = 0

# We place top of cell 50 um below the cortical surface
cell.set_pos(z=top_of_cortex-np.max(cell.zend) - 50)

# Cell bottom needs to be above bottom layer because of convention in
# calculation of ECoG potentials using the Method of Images in
# class RecMEAElectrode. This means that the middle layer
# must extend deeper than the cell.
h = - np.min(cell.zend) + 1

# Synaptic parameters taken from Hendrickson et al 2011
# Excitatory synapse parameters:
synapseParameters_AMPA = {
    'e' : 0,                    #reversal potential
    'syntype' : 'Exp2Syn',      #conductance based exponential synapse
    'tau1' : 1.,                #Time constant, rise
    'tau2' : 3.,                #Time constant, decay
    'weight' : 0.005,           #Synaptic weight
    'record_current' : True,    #record synaptic currents
}
# Excitatory synapse parameters
synapseParameters_NMDA = {         
    'e' : 0,
    'syntype' : 'Exp2Syn',
    'tau1' : 10.,
    'tau2' : 30.,
    'weight' : 0.005,
    'record_current' : True,
}
# Inhibitory synapse parameters
synapseParameters_GABA_A = {         
    'e' : -80,
    'syntype' : 'Exp2Syn',
    'tau1' : 1.,
    'tau2' : 12.,
    'weight' : 0.005,
    'record_current' : True
}
# where to insert, how many, and which input statistics
insert_synapses_AMPA_args = {
    'section' : 'apic',
    'n' : 100,
    'spTimesFun' : LFPy.inputgenerators.get_activation_times_from_distribution,
    'args' : dict(n=1, tstart=cellParameters['tstart'], tstop=cellParameters['tstop'],
                  distribution=scipy.stats.gamma,
                  rvs_args=dict(a=0.5, loc=0., scale=40)
                  )
}
insert_synapses_NMDA_args = {
    'section' : ['dend', 'apic'],
    'n' : 15,
    'spTimesFun' : LFPy.inputgenerators.get_activation_times_from_distribution,
    'args' : dict(n=1, tstart=cellParameters['tstart'], tstop=cellParameters['tstop'],
                  distribution=scipy.stats.gamma,
                  rvs_args=dict(a=2, loc=0, scale=50)
                  )
}
insert_synapses_GABA_A_args = {
    'section' : 'dend',
    'n' : 100,
    'spTimesFun' : LFPy.inputgenerators.get_activation_times_from_distribution,
    'args' : dict(n=1, tstart=cellParameters['tstart'], tstop=cellParameters['tstop'],
                  distribution=scipy.stats.gamma,
                  rvs_args=dict(a=0.5, loc=0., scale=40)
                  )
}
#Insert synapses using the function defined earlier
insert_synapses(synapseParameters_AMPA, **insert_synapses_AMPA_args)
insert_synapses(synapseParameters_NMDA, **insert_synapses_NMDA_args)
insert_synapses(synapseParameters_GABA_A, **insert_synapses_GABA_A_args)

# Conductivities
sigma_S = 1.5  # Condictivity of medium above cortex
sigma_T = 0.3  # Conductivity of cortex

# Define electrode geometry corresponding to a laminar electrode, where contact
# points have a radius r, surface normal vectors N, and LFP calculated as the
# average LFP in n random points on each contact:
num_LFP_elecs = 8

N = np.empty((num_LFP_elecs, 3))
for i in range(N.shape[0]): N[i,] = [1, 0, 0] #normal unit vec. to contacts
# put parameters in dictionary
electrodeParameters = {
    'sigma' : sigma_T,  # Conductivity of cortex (infinite and homogeneous)
    'x' : np.zeros(num_LFP_elecs),      # x,y,z-coordinates of electrode contacts
    'y' : np.zeros(num_LFP_elecs),
    'z' : np.linspace(cell.somapos[2]-200, top_of_cortex, num_LFP_elecs),
    'n' : 20,
    'r' : 10,
    'N' : N,
}

elec_with_MoIParameters = {
    'sigma_S' : sigma_S,  # Conductitivy above cortex
    'sigma_T' : sigma_T,  # Conductivity of cortex
    'sigma_G' : sigma_T,  # Conductivity below cortex. Assume same as cortex
    'h': h,
    'z_shift': -h,
    'x' : np.zeros(num_LFP_elecs),      # x,y,z-coordinates of electrode contacts
    'y' : np.zeros(num_LFP_elecs),
    'z' : np.linspace(cell.somapos[2]-200, top_of_cortex, num_LFP_elecs),
    'n' : 20,
    'r' : 10,
    'N' : N,
    'method': "pointsource",
}

# Define electrode geometry corresponding to an ECoG electrode, where contact
# points have a radius r, surface normal vectors N, and ECoG is calculated as the
# average LFP in n random points on each contact:
num_ECoG_elecs = 1

N = np.empty((num_ECoG_elecs, 3))
for i in range(N.shape[0]): N[i,] = [0, 0, 1] #normal unit vec. to contacts
# put parameters in dictionary

ecogParameters = {
    'sigma_S' : sigma_S,
    'sigma_T' : 0.3,              # Extracellular potential
    'sigma_G' : 0.3,
    'h': h,
    'z_shift': -h,
    'x' : np.zeros(num_ECoG_elecs),      # x,y,z-coordinates of electrode contacts
    'y' : np.zeros(num_ECoG_elecs),
    'z' : np.array([top_of_cortex] * num_ECoG_elecs), # Assume that top of cortex is 50 um above cell top
    'n' : 200,
    'r' : 500, # ECoG radii are often 500-1000 um
    'N' : N,
    'method': "pointsource",
}

#perform NEURON simulation, results saved as attributes in the cell instance
cell.simulate(rec_imem=True)

# Initialize electrode geometry, then calculate the LFP, using the
# electrode classes. Note that now cell is given as input to electrode
# and created after the NEURON simulations are finished
electrode = LFPy.RecExtElectrode(cell, **electrodeParameters)
electrode.calc_lfp()
electrode.LFP -= np.average(electrode.LFP, axis=1)[:, None]

electrode_MoI = LFPy.RecMEAElectrode(cell, **elec_with_MoIParameters)
electrode_MoI.calc_lfp()
electrode_MoI.LFP -= np.average(electrode_MoI.LFP, axis=1)[:, None]

ecog_electrode = LFPy.RecMEAElectrode(cell, **ecogParameters)
ecog_electrode.calc_lfp()
ecog_electrode.LFP -= np.average(ecog_electrode.LFP, axis=1)[:, None]

#plotting some variables and geometry, saving output to .pdf.
fig = plot_LFP_and_ECoG(cell, electrode, electrode_MoI, ecog_electrode)
fig.savefig('example_ECoG.pdf', dpi=300)
plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''plotting script for figure 7 and 8 in manuscript preprint for parallel
performance of LFPy

Copyright (C) 2018 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.'''
import os
import numpy as np
import example_parallel_network_parameterspace as ps
from example_parallel_network_parameters import PSET
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter

plt.rcParams['axes.color_cycle'] = [
    u'#1f77b4',
    u'#1f77b4',
    u'#ff7f0e',
    u'#ff7f0e',
    u'#2ca02c',
    u'#2ca02c',
    u'#d62728',
    u'#d62728',
    ]
plt.rcParams.update({
    'axes.xmargin': 0.0,
    'axes.ymargin': 0.0,
})
fontsize=14
titlesize=16
legendsize=12
plt.rcParams.update({
    'axes.xmargin': 0.0,
    'axes.ymargin': 0.0,
    'axes.labelsize': fontsize,
    'axes.titlesize': titlesize,
    'figure.titlesize': fontsize,
    'font.size': fontsize,
    'legend.fontsize': legendsize,
})

def get_plot_array(x, shape=(3, -1)):
    return np.array(x).reshape(shape)


plt.close('all')

for (PS0, PS1), figname in zip([['MPI', 'POP'], ['MPI5', 'POP5']], ['figure_7', 'figure_8']):

    # fig, axes = plt.subplots(2, 1, sharey=True, figsize=(8, 8))
    fig, axes = plt.subplots(1, 3, sharey=True, figsize=(16, 6))
    fig.subplots_adjust(left=0.05, right=0.85, bottom=0.15, top=0.9)
    
    # plot scaling with MPI pool size and fixed network size
    ax = axes[0]
    
    for ls, COMPUTE_LFP in zip(['-o', ':o'], [True, False]):
        # containers
        MPISIZE = []
        # init = []
        params = []
        pops = []
        conn = []
        sim = []
        # save = []
        ps_ids = []
        
        for pset in ps.PSPACES[PS0].iter_inner():
            if pset.COMPUTE_LFP == COMPUTE_LFP:
                # get identifier
                ps_id = ps.get_unique_id(pset)    
                
                # set reference network size
                PSET.populationParameters['POP_SIZE'] = np.array(pset.POP_SIZE_REF)
                
                # create x-axis
                MPISIZE.append(pset.MPISIZE)
                
                try:
                    # load log file
                    keys = np.loadtxt(os.path.join(ps.OUTPUTDIR, ps_id, 'log.txt'), usecols=[0], dtype=str)
                    values = np.loadtxt(os.path.join(ps.OUTPUTDIR, ps_id, 'log.txt'), usecols=[1])
                
                    # fill in lists, try-excepts for possibly failed simulations
                    data = dict(zip(keys, values))
                    try:
                        params.append(data['parameters'])
                    except KeyError:
                        params.append(np.nan)
                        print('missing param. time id {}'.format(ps_id))
                    try:
                        pops.append(data['population'])
                    except KeyError:
                        pops.append(np.nan)
                        print('missing pop. time id {}'.format(ps_id))
                    try:
                        conn.append(data['connections'])
                    except KeyError:
                        conn.append(np.nan)
                        print('missing conn. time id {}'.format(ps_id))
                    try:
                        sim.append(data['simulation'])
                    except KeyError:
                        sim.append(np.nan)
                        print('missing sim. time id {}'.format(ps_id))

                except IOError:
                    print('fail to load param id {}'.format(ps_id))
                    params.append(np.nan)
                    pops.append(np.nan)
                    conn.append(np.nan)
                    sim.append(np.nan)
                ps_ids.append(ps_id)
                
        shape = (-1, len(ps.PSPACES[PS0]['GLOBALSEED']))
        x = get_plot_array(MPISIZE, shape)[:, 0]
        
        y = get_plot_array(params, shape)
        ax.loglog(x, y.mean(axis=1), ls,
                  label='par.'+', {} E.P'.format('w.' if COMPUTE_LFP else 'w.o.'),
                  lw=2, ms=8,
                  basex=2, basey=2,
                  )
        ax.errorbar(x, y.mean(axis=1), yerr=y.std(axis=1), fmt='none', capsize=4, label='_nolegend_', lw=2, ms=8)
        
        y = get_plot_array(pops, shape)
        ax.loglog(x, y.mean(axis=1), ls,
                  label='pop.'+', {} E.P'.format('w.' if COMPUTE_LFP else 'w.o.'),
                  lw=2, ms=8,
                  basex=2, basey=2,
                  )
        ax.errorbar(x, y.mean(axis=1), yerr=y.std(axis=1), fmt='none', capsize=4, label='_nolegend_', lw=2, ms=8)
        
        y = get_plot_array(conn, shape)
        ax.loglog(x, y.mean(axis=1), ls,
                  label='conn.'+', {} E.P'.format('w.' if COMPUTE_LFP else 'w.o.'),
                  lw=2, ms=8,
                  basex=2, basey=2,
                  )
        ax.errorbar(x, y.mean(axis=1), yerr=y.std(axis=1), fmt='none', capsize=4, label='_nolegend_', lw=2, ms=8)
        
        y = get_plot_array(sim, shape)
        ax.loglog(x, y.mean(axis=1), ls,
                  label='sim.'+', {} E.P'.format('w.' if COMPUTE_LFP else 'w.o.'),
                  lw=2, ms=8,
                  basex=2, basey=2,
                  )
        ax.errorbar(x, y.mean(axis=1), yerr=y.std(axis=1), fmt='none', capsize=4, label='_nolegend_', lw=2, ms=8)
    
    
    
    ax.minorticks_off()
    ax.axis(ax.axis('tight'))
    ax.grid('on', which='major')
    
    
    ax.set_xticks(np.unique(MPISIZE))
    ax.set_xticklabels(np.unique(MPISIZE), rotation=90)
    ax.set_yticks([2.**x for x in range(16)])
    ax.set_yticklabels([2**x for x in range(16)])
    for axis in [ax.xaxis, ax.yaxis]:
        axis.set_major_formatter(ScalarFormatter())
    ax.set_xlabel(r'$N_\mathrm{MPI}$', labelpad=0)
    ax.set_ylabel('time (s)', labelpad=0)
    ax.set_title(r'$N_\mathrm{%s}^{(1)}=%i, N_\mathrm{%s}^{(1)}=%i$' % (PSET.populationParameters['m_type'][0].replace('_', '\_'),
                                                            PSET.populationParameters['POP_SIZE'][0],
                                                            PSET.populationParameters['m_type'][1].replace('_', '\_'),
                                                            PSET.populationParameters['POP_SIZE'][1]))
    ax.text(-0.05, 1.05, 'A',
        horizontalalignment='center',
        verticalalignment='center',
        fontsize=16, fontweight='demibold',
        transform=ax.transAxes)
    axis = ax.axis()
    
    
    # plot scaling with population size and fixed MPI pool size
    for ls, COMPUTE_LFP in zip(['-', ':'], [True, False]):
        for axindex, marker, PRESERVE in zip([1, 2], ['o', 'o'], ['total', 'indegree']):
            ax = axes[axindex]
            # containers
            POPSCALING = []
            # init = []
            params = []
            pops = []
            conn = []
            sim = []
            # save = []
            ps_ids = []
            
            for pset in ps.PSPACES[PS1].iter_inner():
                if pset.COMPUTE_LFP == COMPUTE_LFP and pset.PRESERVE==PRESERVE:
                    # get identifier
                    ps_id = ps.get_unique_id(pset)    
                
                    # create x-axis
                    POPSCALING.append(pset.POPSCALING)
                
                    try:
                        # load log file
                        keys = np.loadtxt(os.path.join(ps.OUTPUTDIR, ps_id, 'log.txt'), usecols=[0], dtype=str)
                        values = np.loadtxt(os.path.join(ps.OUTPUTDIR, ps_id, 'log.txt'), usecols=[1])
                    
                        # fill in lists, try-excepts for possibly failed simulations
                        data = dict(zip(keys, values))    
                        try:
                            params.append(data['parameters'])
                        except KeyError:
                            params.append(np.nan)
                            print('missing param. time id {}'.format(ps_id))
                        try:
                            pops.append(data['population'])
                        except KeyError:
                            pops.append(np.nan)
                            print('missing pop. time id {}'.format(ps_id))
                        try:
                            conn.append(data['connections'])
                        except KeyError:
                            conn.append(np.nan)
                            print('missing conn. time id {}'.format(ps_id))
                        try:
                            sim.append(data['simulation'])
                        except KeyError:
                            sim.append(np.nan)
                            print('missing sim. time id {}'.format(ps_id))
                    except IOError:
                        print('fail to load param id {}'.format(ps_id))
                        params.append(np.nan)
                        pops.append(np.nan)
                        conn.append(np.nan)
                        sim.append(np.nan)
                    ps_ids.append(ps_id)
    
    
            shape = (-1, len(ps.PSPACES[PS1]['GLOBALSEED']))
            x = get_plot_array(POPSCALING, shape)[:, 0]
            
            y = get_plot_array(params, shape)
            ax.loglog(x, y.mean(axis=1), ls, marker=marker,
                      label='par.'+', {} E.P'.format('w.' if COMPUTE_LFP else 'w.o.'),
                      lw=2, ms=8,
                      basex=2, basey=2
                      )
            ax.errorbar(x, y.mean(axis=1), yerr=y.std(axis=1), fmt='none', capsize=4, label='_nolegend_', lw=2, ms=8)
            
            y = get_plot_array(pops, shape)
            ax.loglog(x, y.mean(axis=1), ls, marker=marker,
                      label='pop.'+', {} E.P'.format('w.' if COMPUTE_LFP else 'w.o.'),
                      lw=2, ms=8,
                      basex=2, basey=2
                      )
            ax.errorbar(x, y.mean(axis=1), yerr=y.std(axis=1), fmt='none', capsize=4, label='_nolegend_', lw=2, ms=8)
            
            y = get_plot_array(conn, shape)
            ax.loglog(x, y.mean(axis=1), ls, marker=marker,
                      label='conn.'+', {} E.P'.format('w.' if COMPUTE_LFP else 'w.o.'),
                      lw=2, ms=8,
                      basex=2, basey=2
                      )
            ax.errorbar(x, y.mean(axis=1), yerr=y.std(axis=1), fmt='none', capsize=4, label='_nolegend_', lw=2, ms=8)
            
            y = get_plot_array(sim, shape)
            ax.loglog(x, y.mean(axis=1), ls, marker=marker,
                      label='sim.'+', {} E.P'.format('w.' if COMPUTE_LFP else 'w.o.'),
                      lw=2, ms=8,
                      basex=2, basey=2
                      )
            ax.errorbar(x, y.mean(axis=1), yerr=y.std(axis=1), fmt='none', capsize=4, label='_nolegend_', lw=2, ms=8)
            
        
            ax.axis(ax.axis('tight'))
            ax.grid('on', which='major')
    
        
            ax.set_xticks(np.unique(POPSCALING))
            ax.set_xticklabels(np.unique(POPSCALING), rotation=90)
        
            for axis in [ax.xaxis, ax.yaxis]:
                axis.set_major_formatter(ScalarFormatter())

            ax.set_xlabel(r'relative network size $b$', labelpad=10)
            ax.set_title(r'$N_\mathrm{%s}^{(1)}=%i, N_\mathrm{%s}^{(1)}=%i$,' % (PSET.populationParameters['m_type'][0].replace('_', '\_'),
                                                                    PSET.populationParameters['POP_SIZE'][0],
                                                                    PSET.populationParameters['m_type'][1].replace('_', '\_'),
                                                                    PSET.populationParameters['POP_SIZE'][1],)
                         +'\n' + r'$N_\mathrm{MPI}=%i$, $k_{YX}^{(b)}=%sK_{YX}^{(1)}/N_Y^{(1)}$' % (pset.MPISIZE, 'r' if PRESERVE=='indegree' else ''))
            ax.text(-0.05, 1.05, 'BC'[axindex-1],
                horizontalalignment='center',
                verticalalignment='center',
                fontsize=16, fontweight='demibold',
                transform=ax.transAxes)
            axis = ax.axis()

        ax.legend(bbox_to_anchor=(1.01, 0), loc='lower left', borderaxespad=0.)
    
    fig.savefig(figname + '.pdf', bbox_inches='tight')
